# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags for Workstation Config related commands."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import argparse

from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.command_lib.util.concepts import concept_parsers


def AddAsyncFlag(parser):
  """Adds --async flag."""
  base.ASYNC_FLAG.AddToParser(parser)


def LocationsAttributeConfig(help_text=''):
  """Create a location attribute in resource argument.

  Args:
    help_text: If set, overrides default help text for `--location`

  Returns:
    Location resource argument parameter config
  """
  return concepts.ResourceParameterAttributeConfig(
      name='region',
      help_text=help_text if help_text else
      ('The location for the {resource}.'))


def ClustersAttributeConfig(help_text=''):
  """Create a cluster attribute in resource argument.

  Args:
    help_text: If set, overrides default help text for `--cluster`

  Returns:
    Cluster resource argument parameter config
  """
  return concepts.ResourceParameterAttributeConfig(
      name='cluster',
      help_text=help_text if help_text else ('The cluster for the {resource}.'))


def ConfigsAttributeConfig(help_text=''):
  """Create a config attribute in resource argument.

  Args:
    help_text: If set, overrides default help text for `config`

  Returns:
    Config resource argument parameter config
  """
  return concepts.ResourceParameterAttributeConfig(
      name='config',
      help_text=help_text if help_text else ('The config for the {resource}.'))


def WorkstationsAttributeConfig(help_text=''):
  """Create a workstation attribute in resource argument.

  Args:
    help_text: If set, overrides default help text for `workstation`

  Returns:
    Workstation resource argument parameter config
  """
  return concepts.ResourceParameterAttributeConfig(
      name='workstation',
      help_text=help_text if help_text else ('The workstation.'))


def AddConfigResourceArg(parser, api_version='v1beta'):
  """Create a config resource argument."""
  spec = concepts.ResourceSpec(
      'workstations.projects.locations.workstationClusters.workstationConfigs',
      resource_name='config',
      api_version=api_version,
      workstationConfigsId=ConfigsAttributeConfig(),
      workstationClustersId=ClustersAttributeConfig(),
      locationsId=LocationsAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
  )
  concept_parsers.ConceptParser.ForResource(
      'config',
      spec,
      'The group of arguments defining a config',
      required=True,
  ).AddToParser(parser)


def AddWorkstationResourceArg(parser, api_version='v1beta'):
  """Create a workstation resource argument."""
  spec = concepts.ResourceSpec(
      'workstations.projects.locations.workstationClusters.workstationConfigs.workstations',
      resource_name='workstation',
      api_version=api_version,
      workstationsId=WorkstationsAttributeConfig(),
      workstationConfigsId=ConfigsAttributeConfig(),
      workstationClustersId=ClustersAttributeConfig(),
      locationsId=LocationsAttributeConfig(),
      projectsId=concepts.DEFAULT_PROJECT_ATTRIBUTE_CONFIG,
  )
  concept_parsers.ConceptParser.ForResource(
      'workstation',
      spec,
      'The group of arguments defining a workstation',
      required=True,
  ).AddToParser(parser)


def AddIdleTimeoutFlag(parser, use_default=True):
  """Adds an --idle-timeout flag to the given parser."""
  help_text = """\
  How long (in seconds) to wait before automatically stopping an instance that
  hasn't received any user traffic. A value of 0 indicates that this instance
  should never time out due to idleness.
  """
  parser.add_argument(
      '--idle-timeout',
      default=7200 if use_default else None,
      type=int,
      help=help_text,
  )


def AddRunningTimeoutFlag(parser, use_default=True):
  """Adds an --running-timeout flag to the given parser."""
  help_text = """\
  How long (in seconds) to wait before automatically stopping a workstation
  after it started. A value of 0 indicates that workstations using this config
  should never time out.
  """
  parser.add_argument(
      '--running-timeout',
      default=7200 if use_default else None,
      type=int,
      help=help_text,
  )


def AddMachineTypeFlag(parser, use_default=True):
  """Adds a --machine-type flag to the given parser."""
  help_text = """\
  Machine type determines the specifications of the Compute Engine machines
  that the workstations created under this configuration will run on."""
  parser.add_argument(
      '--machine-type',
      choices=[
          'e2-standard-2',
          'e2-standard-4',
          'e2-standard-8',
          'e2-standard-16',
          'e2-standard-32',
          'n2-standard-8',
          'n1-standard-4',
          't2d-standard-6',
          'n2d-standard-2',
          'n2d-standard-4',
          'n2d-standard-8',
          'n2d-standard-16',
          'n2d-standard-32',
      ],
      default='e2-standard-4' if use_default else None,
      help=help_text,
  )


def AddServiceAccountFlag(parser):
  """Adds a --service-account flag to the given parser."""
  help_text = """\
  Email address of the service account that will be used on VM instances used to
  support this config. This service account must have permission to pull the
  specified container image. If not set, VMs will run without a service account,
  in which case the image must be publicly accessible."""
  parser.add_argument('--service-account', help=help_text)


def AddNetworkTags(parser):
  """Adds a --network-tags flag to the given parser."""
  help_text = """\
  Network tags to add to the Google Compute Engine machines backing the
  Workstations.

  Example:

    $ {command} --network-tags=tag_1,tag_2
  """
  parser.add_argument(
      '--network-tags',
      metavar='NETWORK_TAGS',
      type=arg_parsers.ArgList(),
      help=help_text)


def AddPoolSize(parser, use_default=True):
  """Adds a --pool-size flag to the given parser."""
  help_text = """\
  Number of instances to pool for faster Workstation starup."""
  parser.add_argument(
      '--pool-size',
      default=0 if use_default else None,
      type=int,
      help=help_text,
  )


def AddDisablePublicIpAddresses(parser, use_default=True):
  """Adds a --disable-public-ip-addresses flag to the given parser."""
  help_text = """\
  Default value is false.
  If set, instances will have no public IP address."""
  parser.add_argument(
      '--disable-public-ip-addresses',
      action='store_true',
      default=False if use_default else None,
      help=help_text,
  )


def AddShieldedSecureBoot(parser, use_default=True):
  """Adds --shielded-secure-boot flag to the given parser."""
  help_text = """\
  Default value is false.
  If set, instances will have Secure Boot enabled."""
  parser.add_argument(
      '--shielded-secure-boot',
      action='store_true',
      default=False if use_default else None,
      help=help_text,
  )


def AddShieldedVtpm(parser, use_default=True):
  """Adds a --shielded-vtpm flag to the given parser."""
  help_text = """\
  Default value is false.
  If set, instances will have vTPM enabled."""
  parser.add_argument(
      '--shielded-vtpm',
      action='store_true',
      default=False if use_default else None,
      help=help_text,
  )


def AddShieldedIntegrityMonitoring(parser, use_default=True):
  """Adds a --shielded-integrity-monitoring flag to the given parser."""
  help_text = """\
  Default value is false.
  If set, instances will have integrity monitoring enabled."""
  parser.add_argument(
      '--shielded-integrity-monitoring',
      action='store_true',
      default=False if use_default else None,
      help=help_text,
  )


def AddEnableConfidentialCompute(parser, use_default=True):
  """Adds an --enable-confidential-compute flag to the given parser."""
  help_text = """\
  Default value is false.
  If set, instances will have confidential compute enabled."""
  parser.add_argument(
      '--enable-confidential-compute',
      action='store_true',
      default=False if use_default else None,
      help=help_text,
  )


def AddBootDiskSize(parser, use_default=True):
  """Adds a --boot-disk-size flag to the given parser."""
  help_text = """\
  Size of the boot disk in GB."""
  parser.add_argument(
      '--boot-disk-size',
      default=50 if use_default else None,
      type=int,
      help=help_text,
  )


def AddPdDiskType(parser):
  """Adds a --pd-disk-type flag to the given parser."""
  help_text = """\
  Type of the persistent directory."""
  parser.add_argument(
      '--pd-disk-type',
      choices=['pd-standard', 'pd-balanced', 'pd-ssd'],
      default='pd-standard',
      help=help_text)


def AddPdDiskSize(parser):
  """Adds a --pd-disk-size flag to the given parser."""
  help_text = """\
  Size of the persistent directory in GB."""
  parser.add_argument(
      '--pd-disk-size',
      choices=[10, 50, 100, 200, 500, 1000],
      default=200,
      type=int,
      help=help_text)


def AddPdReclaimPolicy(parser):
  """Adds a --pd-reclaim-policy flag to the given parser."""
  help_text = """\
  What should happen to the disk after the Workstation is deleted."""
  parser.add_argument(
      '--pd-reclaim-policy',
      choices={
          'delete':
              'The persistent disk will be deleted with the Workstation.',
          'retain':
              'The persistent disk will be remain after the workstation is deleted and the administrator must manually delete the disk.'
      },
      default='delete',
      help=help_text)


def AddContainerImageField(parser, use_default=True):
  """Adds the --container-predefined-image and --container-custom-image flags to the given parser.
  """
  predefined_image_help_text = """\
  Code editor on base images."""
  custom_image_help_text = """\
  A docker image for the workstation. This image must be accessible by the
  service account configured in this configuration (--service-account). If no
  service account is defined, this image must be public.
  """
  group = parser.add_mutually_exclusive_group()
  group.add_argument(
      '--container-predefined-image',
      choices={
          'codeoss': 'Code OSS',
          'intellij': 'IntelliJ IDEA Ultimate',
          'pycharm': 'PyCharm Professional',
          'rider': 'Rider',
          'webstorm': 'WebStorm',
          'phpstorm': 'PhpStorm',
          'rubymine': 'RubyMine',
          'goland': 'GoLand',
          'clion': 'CLion',
          'base-image': 'Base image - no IDE',
      },
      default='codeoss' if use_default else None,
      help=predefined_image_help_text,
  )

  group.add_argument(
      '--container-custom-image', type=str, help=custom_image_help_text)


def AddContainerCommandField(parser):
  """Adds a --container-command flag to the given parser."""
  help_text = """\
  If set, overrides the default ENTRYPOINT specified by the image.

  Example:

    $ {command} --container-command=executable,parameter_1,parameter_2"""
  parser.add_argument(
      '--container-command',
      metavar='CONTAINER_COMMAND',
      type=arg_parsers.ArgList(),
      help=help_text)


def AddContainerArgsField(parser):
  """Adds a --container-args flag to the given parser."""
  help_text = """\
  Arguments passed to the entrypoint.

  Example:

    $ {command} --container-args=arg_1,arg_2"""
  parser.add_argument(
      '--container-args',
      metavar='CONTAINER_ARGS',
      type=arg_parsers.ArgList(),
      help=help_text)


def AddContainerEnvField(parser):
  """Adds a --container-env flag to the given parser."""
  help_text = """\
  Environment variables passed to the container.

  Example:

    $ {command} --container-env=key1=value1,key2=value2"""
  parser.add_argument(
      '--container-env',
      metavar='CONTAINER_ENV',
      type=arg_parsers.ArgDict(key_type=str, value_type=str),
      help=help_text)


def AddContainerWorkingDirField(parser):
  """Adds a --container-working-dir flag to the given parser."""
  help_text = """\
  If set, overrides the default DIR specified by the image."""
  parser.add_argument('--container-working-dir', help=help_text)


def AddContainerRunAsUserField(parser):
  """Adds a --container-run-as-user flag to the given parser."""
  help_text = """\
  If set, overrides the USER specified in the image with the given
  uid."""
  parser.add_argument('--container-run-as-user', type=int, help=help_text)


def AddWorkstationPortField(parser):
  """Adds a workstation-port flag to the given parser."""
  help_text = """\
  The port on the workstation to which traffic should be sent."""
  parser.add_argument('workstation_port', type=int, help=help_text)


def AddPortField(parser):
  """Adds a --port flag to the given parser."""
  help_text = """\
  The port on the workstation to which traffic should be sent."""
  parser.add_argument('--port', type=int, default=22, help=help_text)


def AddLocalHostPortField(parser):
  """Adds a --local-host-port flag to the given parser."""
  help_text = """\
  `LOCAL_HOST:LOCAL_PORT` on which gcloud should bind and listen for connections
  that should be tunneled.

  `LOCAL_PORT` may be omitted, in which case it is treated as 0 and an arbitrary
  unused local port is chosen. The colon also may be omitted in that case.

  If `LOCAL_PORT` is 0, an arbitrary unused local port is chosen."""
  parser.add_argument(
      '--local-host-port',
      type=arg_parsers.HostPort.Parse,
      default='localhost:0',
      help=help_text)


def AddCommandField(parser):
  """Adds a --command flag to the given parser."""
  help_text = """\
      A command to run on the workstation.

      Runs the command on the target workstation and then exits.
      """
  parser.add_argument('--command', type=str, help=help_text)


def AddSshArgsAndUserField(parser):
  """Adds a --user flag to the given parser."""
  help_text = """\
  The username with which to SSH.
  """
  parser.add_argument('--user', type=str, default='user', help=help_text)

  help_text = """\
  Flags and positionals passed to the underlying ssh implementation."""
  parser.add_argument('ssh_args', nargs=argparse.REMAINDER, help=help_text)



#!/usr/bin/env python
#
# Copyright 2015 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A helper function that executes a series of List queries for many APIs."""

from apitools.base.py import encoding
import six

__all__ = [
    'YieldFromList',
]


def _GetattrNested(message, attribute):
    """Gets a possibly nested attribute.

    Same as getattr() if attribute is a string;
    if attribute is a tuple, returns the nested attribute referred to by
    the fields in the tuple as if they were a dotted accessor path.

    (ex _GetattrNested(msg, ('foo', 'bar', 'baz')) gets msg.foo.bar.baz
    """
    if isinstance(attribute, six.string_types):
        return getattr(message, attribute)
    elif len(attribute) == 0:
        return message
    else:
        return _GetattrNested(getattr(message, attribute[0]), attribute[1:])


def _SetattrNested(message, attribute, value):
    """Sets a possibly nested attribute.

    Same as setattr() if attribute is a string;
    if attribute is a tuple, sets the nested attribute referred to by
    the fields in the tuple as if they were a dotted accessor path.

    (ex _SetattrNested(msg, ('foo', 'bar', 'baz'), 'v') sets msg.foo.bar.baz
    to 'v'
    """
    if isinstance(attribute, six.string_types):
        return setattr(message, attribute, value)
    elif len(attribute) < 1:
        raise ValueError("Need an attribute to set")
    elif len(attribute) == 1:
        return setattr(message, attribute[0], value)
    else:
        return setattr(_GetattrNested(message, attribute[:-1]),
                       attribute[-1], value)


def YieldFromList(
        service, request, global_params=None, limit=None, batch_size=100,
        method='List', field='items', predicate=None,
        current_token_attribute='pageToken',
        next_token_attribute='nextPageToken',
        batch_size_attribute='maxResults',
        get_field_func=_GetattrNested):
    """Make a series of List requests, keeping track of page tokens.

    Args:
      service: apitools_base.BaseApiService, A service with a .List() method.
      request: protorpc.messages.Message, The request message
          corresponding to the service's .List() method, with all the
          attributes populated except the .maxResults and .pageToken
          attributes.
      global_params: protorpc.messages.Message, The global query parameters to
           provide when calling the given method.
      limit: int, The maximum number of records to yield. None if all available
          records should be yielded.
      batch_size: int, The number of items to retrieve per request.
      method: str, The name of the method used to fetch resources.
      field: str, The field in the response that will be a list of items.
      predicate: lambda, A function that returns true for items to be yielded.
      current_token_attribute: str or tuple, The name of the attribute in a
          request message holding the page token for the page being
          requested. If a tuple, path to attribute.
      next_token_attribute: str or tuple, The name of the attribute in a
          response message holding the page token for the next page. If a
          tuple, path to the attribute.
      batch_size_attribute: str or tuple, The name of the attribute in a
          request message holding the maximum number of results to be
          returned. None if caller-specified batch size is unsupported.
          If a tuple, path to the attribute.
      get_field_func: Function that returns the items to be yielded. Argument
          is response message, and field.

    Yields:
      protorpc.message.Message, The resources listed by the service.

    """
    request = encoding.CopyProtoMessage(request)
    _SetattrNested(request, current_token_attribute, None)
    while limit is None or limit:
        if batch_size_attribute:
            # On Py3, None is not comparable so min() below will fail.
            # On Py2, None is always less than any number so if batch_size
            # is None, the request_batch_size will always be None regardless
            # of the value of limit. This doesn't generally strike me as the
            # correct behavior, but this change preserves the existing Py2
            # behavior on Py3.
            if batch_size is None:
                request_batch_size = None
            else:
                request_batch_size = min(batch_size, limit or batch_size)
            _SetattrNested(request, batch_size_attribute, request_batch_size)
        response = getattr(service, method)(request,
                                            global_params=global_params)
        items = get_field_func(response, field)
        if predicate:
            items = list(filter(predicate, items))
        for item in items:
            yield item
            if limit is None:
                continue
            limit -= 1
            if not limit:
                return
        token = _GetattrNested(response, next_token_attribute)
        if not token:
            return
        _SetattrNested(request, current_token_attribute, token)



#!/usr/bin/env python
# Copyright 2016 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OAuth 2.0 Credentials.

This module provides credentials based on OAuth 2.0 access and refresh tokens.
These credentials usually access resources on behalf of a user (resource
owner).

Specifically, this is intended to use access tokens acquired using the
`Authorization Code grant`_ and can refresh those tokens using a
optional `refresh token`_.

Obtaining the initial access and refresh token is outside of the scope of this
module. Consult `rfc6749 section 4.1`_ for complete details on the
Authorization Code grant flow.

.. _Authorization Code grant: https://tools.ietf.org/html/rfc6749#section-1.3.1
.. _refresh token: https://tools.ietf.org/html/rfc6749#section-6
.. _rfc6749 section 4.1: https://tools.ietf.org/html/rfc6749#section-4.1
"""

from datetime import datetime
import io
import json
import logging

import six

from google.auth import _cloud_sdk
from google.auth import _helpers
from google.auth import credentials
from google.auth import exceptions
from google.oauth2 import reauth

_LOGGER = logging.getLogger(__name__)


# The Google OAuth 2.0 token endpoint. Used for authorized user credentials.
_GOOGLE_OAUTH2_TOKEN_ENDPOINT = "https://oauth2.googleapis.com/token"


class Credentials(credentials.ReadOnlyScoped, credentials.CredentialsWithQuotaProject):
    """Credentials using OAuth 2.0 access and refresh tokens.

    The credentials are considered immutable. If you want to modify the
    quota project, use :meth:`with_quota_project` or ::

        credentials = credentials.with_quota_project('myproject-123')

    Reauth is disabled by default. To enable reauth, set the
    `enable_reauth_refresh` parameter to True in the constructor. Note that
    reauth feature is intended for gcloud to use only.
    If reauth is enabled, `pyu2f` dependency has to be installed in order to use security
    key reauth feature. Dependency can be installed via `pip install pyu2f` or `pip install
    google-auth[reauth]`.
    """

    def __init__(
        self,
        token,
        refresh_token=None,
        id_token=None,
        token_uri=None,
        client_id=None,
        client_secret=None,
        scopes=None,
        default_scopes=None,
        quota_project_id=None,
        expiry=None,
        rapt_token=None,
        refresh_handler=None,
        enable_reauth_refresh=False,
        granted_scopes=None,
    ):
        """
        Args:
            token (Optional(str)): The OAuth 2.0 access token. Can be None
                if refresh information is provided.
            refresh_token (str): The OAuth 2.0 refresh token. If specified,
                credentials can be refreshed.
            id_token (str): The Open ID Connect ID Token.
            token_uri (str): The OAuth 2.0 authorization server's token
                endpoint URI. Must be specified for refresh, can be left as
                None if the token can not be refreshed.
            client_id (str): The OAuth 2.0 client ID. Must be specified for
                refresh, can be left as None if the token can not be refreshed.
            client_secret(str): The OAuth 2.0 client secret. Must be specified
                for refresh, can be left as None if the token can not be
                refreshed.
            scopes (Sequence[str]): The scopes used to obtain authorization.
                This parameter is used by :meth:`has_scopes`. OAuth 2.0
                credentials can not request additional scopes after
                authorization. The scopes must be derivable from the refresh
                token if refresh information is provided (e.g. The refresh
                token scopes are a superset of this or contain a wild card
                scope like 'https://www.googleapis.com/auth/any-api').
            default_scopes (Sequence[str]): Default scopes passed by a
                Google client library. Use 'scopes' for user-defined scopes.
            quota_project_id (Optional[str]): The project ID used for quota and billing.
                This project may be different from the project used to
                create the credentials.
            rapt_token (Optional[str]): The reauth Proof Token.
            refresh_handler (Optional[Callable[[google.auth.transport.Request, Sequence[str]], [str, datetime]]]):
                A callable which takes in the HTTP request callable and the list of
                OAuth scopes and when called returns an access token string for the
                requested scopes and its expiry datetime. This is useful when no
                refresh tokens are provided and tokens are obtained by calling
                some external process on demand. It is particularly useful for
                retrieving downscoped tokens from a token broker.
            enable_reauth_refresh (Optional[bool]): Whether reauth refresh flow
                should be used. This flag is for gcloud to use only.
            granted_scopes (Optional[Sequence[str]]): The scopes that were consented/granted by the user.
                This could be different from the requested scopes and it could be empty if granted
                and requested scopes were same.
        """
        super(Credentials, self).__init__()
        self.token = token
        self.expiry = expiry
        self._refresh_token = refresh_token
        self._id_token = id_token
        self._scopes = scopes
        self._default_scopes = default_scopes
        self._granted_scopes = granted_scopes
        self._token_uri = token_uri
        self._client_id = client_id
        self._client_secret = client_secret
        self._quota_project_id = quota_project_id
        self._rapt_token = rapt_token
        self.refresh_handler = refresh_handler
        self._enable_reauth_refresh = enable_reauth_refresh

    def __getstate__(self):
        """A __getstate__ method must exist for the __setstate__ to be called
        This is identical to the default implementation.
        See https://docs.python.org/3.7/library/pickle.html#object.__setstate__
        """
        state_dict = self.__dict__.copy()
        # Remove _refresh_handler function as there are limitations pickling and
        # unpickling certain callables (lambda, functools.partial instances)
        # because they need to be importable.
        # Instead, the refresh_handler setter should be used to repopulate this.
        del state_dict["_refresh_handler"]
        return state_dict

    def __setstate__(self, d):
        """Credentials pickled with older versions of the class do not have
        all the attributes."""
        self.token = d.get("token")
        self.expiry = d.get("expiry")
        self._refresh_token = d.get("_refresh_token")
        self._id_token = d.get("_id_token")
        self._scopes = d.get("_scopes")
        self._default_scopes = d.get("_default_scopes")
        self._granted_scopes = d.get("_granted_scopes")
        self._token_uri = d.get("_token_uri")
        self._client_id = d.get("_client_id")
        self._client_secret = d.get("_client_secret")
        self._quota_project_id = d.get("_quota_project_id")
        self._rapt_token = d.get("_rapt_token")
        self._enable_reauth_refresh = d.get("_enable_reauth_refresh")
        # The refresh_handler setter should be used to repopulate this.
        self._refresh_handler = None

    @property
    def refresh_token(self):
        """Optional[str]: The OAuth 2.0 refresh token."""
        return self._refresh_token

    @property
    def scopes(self):
        """Optional[str]: The OAuth 2.0 permission scopes."""
        return self._scopes

    @property
    def granted_scopes(self):
        """Optional[Sequence[str]]: The OAuth 2.0 permission scopes that were granted by the user."""
        return self._granted_scopes

    @property
    def token_uri(self):
        """Optional[str]: The OAuth 2.0 authorization server's token endpoint
        URI."""
        return self._token_uri

    @property
    def id_token(self):
        """Optional[str]: The Open ID Connect ID Token.

        Depending on the authorization server and the scopes requested, this
        may be populated when credentials are obtained and updated when
        :meth:`refresh` is called. This token is a JWT. It can be verified
        and decoded using :func:`google.oauth2.id_token.verify_oauth2_token`.
        """
        return self._id_token

    @property
    def client_id(self):
        """Optional[str]: The OAuth 2.0 client ID."""
        return self._client_id

    @property
    def client_secret(self):
        """Optional[str]: The OAuth 2.0 client secret."""
        return self._client_secret

    @property
    def requires_scopes(self):
        """False: OAuth 2.0 credentials have their scopes set when
        the initial token is requested and can not be changed."""
        return False

    @property
    def rapt_token(self):
        """Optional[str]: The reauth Proof Token."""
        return self._rapt_token

    @property
    def refresh_handler(self):
        """Returns the refresh handler if available.

        Returns:
           Optional[Callable[[google.auth.transport.Request, Sequence[str]], [str, datetime]]]:
               The current refresh handler.
        """
        return self._refresh_handler

    @refresh_handler.setter
    def refresh_handler(self, value):
        """Updates the current refresh handler.

        Args:
            value (Optional[Callable[[google.auth.transport.Request, Sequence[str]], [str, datetime]]]):
                The updated value of the refresh handler.

        Raises:
            TypeError: If the value is not a callable or None.
        """
        if not callable(value) and value is not None:
            raise TypeError("The provided refresh_handler is not a callable or None.")
        self._refresh_handler = value

    @_helpers.copy_docstring(credentials.CredentialsWithQuotaProject)
    def with_quota_project(self, quota_project_id):

        return self.__class__(
            self.token,
            refresh_token=self.refresh_token,
            id_token=self.id_token,
            token_uri=self.token_uri,
            client_id=self.client_id,
            client_secret=self.client_secret,
            scopes=self.scopes,
            default_scopes=self.default_scopes,
            granted_scopes=self.granted_scopes,
            quota_project_id=quota_project_id,
            rapt_token=self.rapt_token,
            enable_reauth_refresh=self._enable_reauth_refresh,
        )

    @_helpers.copy_docstring(credentials.CredentialsWithTokenUri)
    def with_token_uri(self, token_uri):

        return self.__class__(
            self.token,
            refresh_token=self.refresh_token,
            id_token=self.id_token,
            token_uri=token_uri,
            client_id=self.client_id,
            client_secret=self.client_secret,
            scopes=self.scopes,
            default_scopes=self.default_scopes,
            granted_scopes=self.granted_scopes,
            quota_project_id=self.quota_project_id,
            rapt_token=self.rapt_token,
            enable_reauth_refresh=self._enable_reauth_refresh,
        )

    @_helpers.copy_docstring(credentials.Credentials)
    def refresh(self, request):
        scopes = self._scopes if self._scopes is not None else self._default_scopes
        # Use refresh handler if available and no refresh token is
        # available. This is useful in general when tokens are obtained by calling
        # some external process on demand. It is particularly useful for retrieving
        # downscoped tokens from a token broker.
        if self._refresh_token is None and self.refresh_handler:
            token, expiry = self.refresh_handler(request, scopes=scopes)
            # Validate returned data.
            if not isinstance(token, six.string_types):
                raise exceptions.RefreshError(
                    "The refresh_handler returned token is not a string."
                )
            if not isinstance(expiry, datetime):
                raise exceptions.RefreshError(
                    "The refresh_handler returned expiry is not a datetime object."
                )
            if _helpers.utcnow() >= expiry - _helpers.REFRESH_THRESHOLD:
                raise exceptions.RefreshError(
                    "The credentials returned by the refresh_handler are "
                    "already expired."
                )
            self.token = token
            self.expiry = expiry
            return

        if (
            self._refresh_token is None
            or self._token_uri is None
            or self._client_id is None
            or self._client_secret is None
        ):
            raise exceptions.RefreshError(
                "The credentials do not contain the necessary fields need to "
                "refresh the access token. You must specify refresh_token, "
                "token_uri, client_id, and client_secret."
            )

        (
            access_token,
            refresh_token,
            expiry,
            grant_response,
            rapt_token,
        ) = reauth.refresh_grant(
            request,
            self._token_uri,
            self._refresh_token,
            self._client_id,
            self._client_secret,
            scopes=scopes,
            rapt_token=self._rapt_token,
            enable_reauth_refresh=self._enable_reauth_refresh,
        )

        self.token = access_token
        self.expiry = expiry
        self._refresh_token = refresh_token
        self._id_token = grant_response.get("id_token")
        self._rapt_token = rapt_token

        if scopes and "scope" in grant_response:
            requested_scopes = frozenset(scopes)
            self._granted_scopes = grant_response["scope"].split()
            granted_scopes = frozenset(self._granted_scopes)
            scopes_requested_but_not_granted = requested_scopes - granted_scopes
            if scopes_requested_but_not_granted:
                # User might be presented with unbundled scopes at the time of
                # consent. So it is a valid scenario to not have all the requested
                # scopes as part of granted scopes but log a warning in case the
                # developer wants to debug the scenario.
                _LOGGER.warning(
                    "Not all requested scopes were granted by the "
                    "authorization server, missing scopes {}.".format(
                        ", ".join(scopes_requested_but_not_granted)
                    )
                )

    @classmethod
    def from_authorized_user_info(cls, info, scopes=None):
        """Creates a Credentials instance from parsed authorized user info.

        Args:
            info (Mapping[str, str]): The authorized user info in Google
                format.
            scopes (Sequence[str]): Optional list of scopes to include in the
                credentials.

        Returns:
            google.oauth2.credentials.Credentials: The constructed
                credentials.

        Raises:
            ValueError: If the info is not in the expected format.
        """
        keys_needed = set(("refresh_token", "client_id", "client_secret"))
        missing = keys_needed.difference(six.iterkeys(info))

        if missing:
            raise ValueError(
                "Authorized user info was not in the expected format, missing "
                "fields {}.".format(", ".join(missing))
            )

        # access token expiry (datetime obj); auto-expire if not saved
        expiry = info.get("expiry")
        if expiry:
            expiry = datetime.strptime(
                expiry.rstrip("Z").split(".")[0], "%Y-%m-%dT%H:%M:%S"
            )
        else:
            expiry = _helpers.utcnow() - _helpers.REFRESH_THRESHOLD

        # process scopes, which needs to be a seq
        if scopes is None and "scopes" in info:
            scopes = info.get("scopes")
            if isinstance(scopes, six.string_types):
                scopes = scopes.split(" ")

        return cls(
            token=info.get("token"),
            refresh_token=info.get("refresh_token"),
            token_uri=_GOOGLE_OAUTH2_TOKEN_ENDPOINT,  # always overrides
            scopes=scopes,
            client_id=info.get("client_id"),
            client_secret=info.get("client_secret"),
            quota_project_id=info.get("quota_project_id"),  # may not exist
            expiry=expiry,
            rapt_token=info.get("rapt_token"),  # may not exist
        )

    @classmethod
    def from_authorized_user_file(cls, filename, scopes=None):
        """Creates a Credentials instance from an authorized user json file.

        Args:
            filename (str): The path to the authorized user json file.
            scopes (Sequence[str]): Optional list of scopes to include in the
                credentials.

        Returns:
            google.oauth2.credentials.Credentials: The constructed
                credentials.

        Raises:
            ValueError: If the file is not in the expected format.
        """
        with io.open(filename, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)
            return cls.from_authorized_user_info(data, scopes)

    def to_json(self, strip=None):
        """Utility function that creates a JSON representation of a Credentials
        object.

        Args:
            strip (Sequence[str]): Optional list of members to exclude from the
                                   generated JSON.

        Returns:
            str: A JSON representation of this instance. When converted into
            a dictionary, it can be passed to from_authorized_user_info()
            to create a new credential instance.
        """
        prep = {
            "token": self.token,
            "refresh_token": self.refresh_token,
            "token_uri": self.token_uri,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "scopes": self.scopes,
            "rapt_token": self.rapt_token,
        }
        if self.expiry:  # flatten expiry timestamp
            prep["expiry"] = self.expiry.isoformat() + "Z"

        # Remove empty entries (those which are None)
        prep = {k: v for k, v in prep.items() if v is not None}

        # Remove entries that explicitely need to be removed
        if strip is not None:
            prep = {k: v for k, v in prep.items() if k not in strip}

        return json.dumps(prep)


class UserAccessTokenCredentials(credentials.CredentialsWithQuotaProject):
    """Access token credentials for user account.

    Obtain the access token for a given user account or the current active
    user account with the ``gcloud auth print-access-token`` command.

    Args:
        account (Optional[str]): Account to get the access token for. If not
            specified, the current active account will be used.
        quota_project_id (Optional[str]): The project ID used for quota
            and billing.
    """

    def __init__(self, account=None, quota_project_id=None):
        super(UserAccessTokenCredentials, self).__init__()
        self._account = account
        self._quota_project_id = quota_project_id

    def with_account(self, account):
        """Create a new instance with the given account.

        Args:
            account (str): Account to get the access token for.

        Returns:
            google.oauth2.credentials.UserAccessTokenCredentials: The created
                credentials with the given account.
        """
        return self.__class__(account=account, quota_project_id=self._quota_project_id)

    @_helpers.copy_docstring(credentials.CredentialsWithQuotaProject)
    def with_quota_project(self, quota_project_id):
        return self.__class__(account=self._account, quota_project_id=quota_project_id)

    def refresh(self, request):
        """Refreshes the access token.

        Args:
            request (google.auth.transport.Request): This argument is required
                by the base class interface but not used in this implementation,
                so just set it to `None`.

        Raises:
            google.auth.exceptions.UserAccessTokenError: If the access token
                refresh failed.
        """
        self.token = _cloud_sdk.get_auth_access_token(self._account)

    @_helpers.copy_docstring(credentials.Credentials)
    def before_request(self, request, method, url, headers):
        self.refresh(request)
        self.apply(headers)



# -*- coding: utf-8 -*- #
# Copyright 2014 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for deleting target HTTPS proxies."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute import utils
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute import scope as compute_scope
from googlecloudsdk.command_lib.compute.target_https_proxies import flags
from googlecloudsdk.command_lib.compute.target_https_proxies import target_https_proxies_utils


def _DetailedHelp():
  return {
      'brief':
          'Delete target HTTPS proxies.',
      'DESCRIPTION':
          """\
      *{command}* deletes one or more target HTTPS proxies.
      """,
      'EXAMPLES':
          """\
      Delete a global target HTTPS proxy by running:

        $ {command} PROXY_NAME

      Delete a regional target HTTPS proxy by running:

        $ {command} PROXY_NAME --region=REGION_NAME
      """,
  }


def _Run(args, holder, target_https_proxy_arg):
  """Issues requests necessary to delete Target HTTPS Proxies."""
  client = holder.client

  target_https_proxy_refs = target_https_proxy_arg.ResolveAsResource(
      args,
      holder.resources,
      default_scope=compute_scope.ScopeEnum.GLOBAL,
      scope_lister=compute_flags.GetDefaultScopeLister(client))

  utils.PromptForDeletion(target_https_proxy_refs)

  requests = []
  for target_https_proxy_ref in target_https_proxy_refs:
    if target_https_proxies_utils.IsRegionalTargetHttpsProxiesRef(
        target_https_proxy_ref):
      requests.append(
          (client.apitools_client.regionTargetHttpsProxies, 'Delete',
           client.messages.ComputeRegionTargetHttpsProxiesDeleteRequest(
               **target_https_proxy_ref.AsDict())))
    else:
      requests.append((client.apitools_client.targetHttpsProxies, 'Delete',
                       client.messages.ComputeTargetHttpsProxiesDeleteRequest(
                           **target_https_proxy_ref.AsDict())))

  return client.MakeRequests(requests)


@base.ReleaseTracks(base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA,
                    base.ReleaseTrack.GA)
class Delete(base.DeleteCommand):
  """Delete target HTTPS proxies."""

  TARGET_HTTPS_PROXY_ARG = None
  detailed_help = _DetailedHelp()

  @classmethod
  def Args(cls, parser):
    cls.TARGET_HTTPS_PROXY_ARG = flags.TargetHttpsProxyArgument(plural=True)
    cls.TARGET_HTTPS_PROXY_ARG.AddArgument(parser, operation_type='delete')
    parser.display_info.AddCacheUpdater(flags.TargetHttpsProxiesCompleter)

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    return _Run(args, holder, self.TARGET_HTTPS_PROXY_ARG)



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""'vmware private-clouds subnets update' command."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.vmware.private_clouds.subnets import SubnetsClient
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.vmware import flags
from googlecloudsdk.core import log

DETAILED_HELP = {
    'DESCRIPTION':
        """
          Update a Subnet. Only ip-cidr-range can be updated. This is a synchronous command and doesn't support `--async` and `--no-async` flags.
        """,
    'EXAMPLES':
        """
        To update a subnet named `my-subnet`, that belongs to the private cloud `my-private-cloud` in project `my-project` and zone `us-west1-a` by changing its ip-cidr-range to `10.0.0.0/24`, run:

          $ {command} my-subnet --private-cloud=my-private-cloud --location=us-west1 --project=my-project --ip-cidr-range=10.0.0.0/24

        Or:

          $ {command} my-subnet --private-cloud=my-private-cloud --ip-cidr-range=10.0.0.0/24

        In the second example, the project and location are taken from gcloud properties `core/project` and `compute/zone`, respectively.
        """
}


@base.Hidden
@base.ReleaseTracks(base.ReleaseTrack.GA)
class Update(base.UpdateCommand):
  """Update a subnet."""

  detailed_help = DETAILED_HELP

  @staticmethod
  def Args(parser):
    """Register flags for this command."""
    flags.AddSubnetArgToParser(parser)
    parser.add_argument(
        '--ip-cidr-range',
        required=True,
        help="""\
        Updated IP CIDR range for this subnet.
        """)

  def Run(self, args):
    subnet = args.CONCEPTS.subnet.Parse()
    client = SubnetsClient()
    operation = client.Update(subnet, args.ip_cidr_range)
    # Since this is a passthrough API, it doesn't return a standard operation.
    # It returns an operation with only two fields: `done` and `response`. If
    # `operation.done == true` we extract the resource from `operation.response`
    # field otherwise we wait for the operation to be completed.
    if operation.done:
      resource = client.GetResponse(operation)
    else:
      resource = client.WaitForOperation(
          operation_ref=client.GetOperationRef(operation),
          message='waiting for subnet [{}] to be updated'.format(
              subnet.RelativeName()),
          has_result=True)
    log.UpdatedResource(resource, kind='subnet', is_async=False)
    return resource



"""Generated client library for policytroubleshooter version v3alpha."""
# NOTE: This file is autogenerated and should not be edited by hand.

from __future__ import absolute_import

from apitools.base.py import base_api
from googlecloudsdk.generated_clients.apis.policytroubleshooter.v3alpha import policytroubleshooter_v3alpha_messages as messages


class PolicytroubleshooterV3alpha(base_api.BaseApiClient):
  """Generated client library for service policytroubleshooter version v3alpha."""

  MESSAGES_MODULE = messages
  BASE_URL = 'https://policytroubleshooter.googleapis.com/'
  MTLS_BASE_URL = 'https://policytroubleshooter.mtls.googleapis.com/'

  _PACKAGE = 'policytroubleshooter'
  _SCOPES = ['https://www.googleapis.com/auth/cloud-platform']
  _VERSION = 'v3alpha'
  _CLIENT_ID = 'CLIENT_ID'
  _CLIENT_SECRET = 'CLIENT_SECRET'
  _USER_AGENT = 'google-cloud-sdk'
  _CLIENT_CLASS_NAME = 'PolicytroubleshooterV3alpha'
  _URL_VERSION = 'v3alpha'
  _API_KEY = None

  def __init__(self, url='', credentials=None,
               get_credentials=True, http=None, model=None,
               log_request=False, log_response=False,
               credentials_args=None, default_global_params=None,
               additional_http_headers=None, response_encoding=None):
    """Create a new policytroubleshooter handle."""
    url = url or self.BASE_URL
    super(PolicytroubleshooterV3alpha, self).__init__(
        url, credentials=credentials,
        get_credentials=get_credentials, http=http, model=model,
        log_request=log_request, log_response=log_response,
        credentials_args=credentials_args,
        default_global_params=default_global_params,
        additional_http_headers=additional_http_headers,
        response_encoding=response_encoding)
    self.iam = self.IamService(self)

  class IamService(base_api.BaseApiService):
    """Service class for the iam resource."""

    _NAME = 'iam'

    def __init__(self, client):
      super(PolicytroubleshooterV3alpha.IamService, self).__init__(client)
      self._upload_configs = {
          }

    def Troubleshoot(self, request, global_params=None):
      r"""Checks whether a member has a specific permission for a specific resource, and explains why the member does or does not have that permission.

      Args:
        request: (GoogleCloudPolicytroubleshooterIamV3alphaTroubleshootIamPolicyRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (GoogleCloudPolicytroubleshooterIamV3alphaTroubleshootIamPolicyResponse) The response message.
      """
      config = self.GetMethodConfig('Troubleshoot')
      return self._RunMethod(
          config, request, global_params=global_params)

    Troubleshoot.method_config = lambda: base_api.ApiMethodInfo(
        http_method='POST',
        method_id='policytroubleshooter.iam.troubleshoot',
        ordered_params=[],
        path_params=[],
        query_params=[],
        relative_path='v3alpha/iam:troubleshoot',
        request_field='<request>',
        request_type_name='GoogleCloudPolicytroubleshooterIamV3alphaTroubleshootIamPolicyRequest',
        response_type_name='GoogleCloudPolicytroubleshooterIamV3alphaTroubleshootIamPolicyResponse',
        supports_download=False,
    )



# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility for updating Memorystore Redis instances."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from apitools.base.py import encoding
from googlecloudsdk.command_lib.redis import util
from googlecloudsdk.core import exceptions
from googlecloudsdk.core.console import console_io
from six.moves import filter  # pylint: disable=redefined-builtin


class NoFieldsSpecified(exceptions.Error):
  """Error for calling update command with no args that represent fields."""


def CheckFieldsSpecifiedGA(unused_instance_ref, args, patch_request):
  """Checks if fields to update are registered for GA track."""
  return CheckFieldsSpecifiedCommon(args, patch_request, [])


def CheckFieldsSpecifiedBeta(unused_instance_ref, args, patch_request):
  """Checks if fields to update are registered for BETA track."""
  additional_update_args = [
      'maintenance_version',
  ]
  return CheckFieldsSpecifiedCommon(args, patch_request, additional_update_args)


def CheckFieldsSpecifiedAlpha(unused_instance_ref, args, patch_request):
  """Checks if fields to update are registered for ALPHA track."""
  additional_update_args = [
      'maintenance_version',
  ]
  return CheckFieldsSpecifiedCommon(args, patch_request, additional_update_args)


def CheckFieldsSpecifiedCommon(args, patch_request, additional_update_args):
  """Checks fields to update that are registered for all tracks."""
  update_args = [
      'clear_labels',
      'display_name',
      'enable_auth',
      'remove_labels',
      'remove_redis_config',
      'size',
      'update_labels',
      'update_redis_config',
      'read_replicas_mode',
      'secondary_ip_range',
      'replica_count',
      'persistence_mode',
      'rdb_snapshot_period',
      'rdb_snapshot_start_time',
      'maintenance_window_day',
      'maintenance_window_hour',
      'maintenance_window_any',
  ] + additional_update_args
  if list(filter(args.IsSpecified, update_args)):
    return patch_request
  raise NoFieldsSpecified(
      'Must specify at least one valid instance parameter to update')


def AddFieldToUpdateMask(field, patch_request):
  update_mask = patch_request.updateMask
  if update_mask:
    if update_mask.count(field) == 0:
      patch_request.updateMask = update_mask + ',' + field
  else:
    patch_request.updateMask = field
  return patch_request


def AddDisplayName(unused_instance_ref, args, patch_request):
  if args.IsSpecified('display_name'):
    patch_request.instance.displayName = args.display_name
    patch_request = AddFieldToUpdateMask('display_name', patch_request)
  return patch_request


def _WarnForDestructiveSizeUpdate(instance_ref, instance):
  """Adds prompt that warns about a destructive size update."""
  messages = util.GetMessagesForResource(instance_ref)
  message = 'Change to instance size requested. '
  if instance.tier == messages.Instance.TierValueValuesEnum.BASIC:
    message += ('Scaling a Basic Tier instance will result in a full cache '
                'flush, and the instance will be unavailable during the '
                'operation. ')
  elif instance.tier == messages.Instance.TierValueValuesEnum.STANDARD_HA:
    message += ('Scaling a Standard Tier instance may result in the loss of '
                'unreplicated data, and the instance will be briefly '
                'unavailable during failover. ')
  else:
    # To future proof this against new instance types, add a default message.
    message += ('Scaling a redis instance may result in data loss, and the '
                'instance will be briefly unavailable during scaling. ')
  message += (
      'For more information please take a look at '
      'https://cloud.google.com/memorystore/docs/redis/scaling-instances')

  console_io.PromptContinue(
      message=message,
      prompt_string='Do you want to proceed with update?',
      cancel_on_no=True)


def AddSize(instance_ref, args, patch_request):
  """Python hook to add size update to the redis instance update request."""
  if args.IsSpecified('size'):
    # Changing size is destructive and users should be warned before proceeding.
    _WarnForDestructiveSizeUpdate(instance_ref, patch_request.instance)
    patch_request.instance.memorySizeGb = args.size
    patch_request = AddFieldToUpdateMask('memory_size_gb', patch_request)
  return patch_request


def RemoveRedisConfigs(instance_ref, args, patch_request):
  if not getattr(patch_request.instance, 'redisConfigs', None):
    return patch_request
  if args.IsSpecified('remove_redis_config'):
    config_dict = encoding.MessageToDict(patch_request.instance.redisConfigs)
    for removed_key in args.remove_redis_config:
      config_dict.pop(removed_key, None)
    patch_request = AddNewRedisConfigs(instance_ref, config_dict, patch_request)
  return patch_request


def UpdateRedisConfigs(instance_ref, args, patch_request):
  if args.IsSpecified('update_redis_config'):
    config_dict = {}
    if getattr(patch_request.instance, 'redisConfigs', None):
      config_dict = encoding.MessageToDict(patch_request.instance.redisConfigs)
    config_dict.update(args.update_redis_config)
    patch_request = AddNewRedisConfigs(instance_ref, config_dict, patch_request)
  return patch_request


def AddNewRedisConfigs(instance_ref, redis_configs_dict, patch_request):
  messages = util.GetMessagesForResource(instance_ref)
  new_redis_configs = util.PackageInstanceRedisConfig(redis_configs_dict,
                                                      messages)
  patch_request.instance.redisConfigs = new_redis_configs
  patch_request = AddFieldToUpdateMask('redis_configs', patch_request)
  return patch_request


def UpdateAuthEnabled(unused_instance_ref, args, patch_request):
  """Hook to add auth_enabled to the update mask of the request."""
  if args.IsSpecified('enable_auth'):
    util.WarnOnAuthEnabled(args.enable_auth)
    patch_request = AddFieldToUpdateMask('auth_enabled', patch_request)
  return patch_request


def UpdateMaintenanceWindowDay(unused_instance_ref, args, patch_request):
  """Hook to update maintenance window day to the update mask of the request."""
  if args.IsSpecified('maintenance_window_day'):
    patch_request = AddFieldToUpdateMask('maintenance_policy', patch_request)
  return patch_request


def UpdateMaintenanceWindowHour(unused_instance_ref, args, patch_request):
  """Hook to update maintenance window hour to the update mask of the request."""
  if args.IsSpecified('maintenance_window_hour'):
    patch_request = AddFieldToUpdateMask('maintenance_policy', patch_request)
  return patch_request


def UpdateMaintenanceWindowAny(unused_instance_ref, args, patch_request):
  """Hook to remove maintenance window."""
  if args.IsSpecified('maintenance_window_any'):
    patch_request.instance.maintenancePolicy = None
    patch_request = AddFieldToUpdateMask('maintenance_policy', patch_request)
  return patch_request


def UpdatePersistenceMode(unused_instance_ref, args, patch_request):
  """Hook to update persistence mode."""
  if args.IsSpecified('persistence_mode'):
    patch_request = AddFieldToUpdateMask('persistence_config', patch_request)
  return patch_request


def UpdateRdbSnapshotPeriod(unused_instance_ref, args, patch_request):
  """Hook to update RDB snapshot period."""
  if args.IsSpecified('rdb_snapshot_period'):
    patch_request = AddFieldToUpdateMask('persistence_config', patch_request)
  return patch_request


def UpdateRdbSnapshotStartTime(unused_instance_ref, args, patch_request):
  """Hook to update RDB snapshot start time."""
  if args.IsSpecified('rdb_snapshot_start_time'):
    patch_request = AddFieldToUpdateMask('persistence_config', patch_request)
  return patch_request


def UpdateReplicaCount(unused_instance_ref, args, patch_request):
  """Hook to update replica count."""
  if args.IsSpecified('replica_count'):
    patch_request = AddFieldToUpdateMask('replica_count', patch_request)
  return patch_request


def UpdateReadReplicasMode(unused_instance_ref, args, patch_request):
  """Hook to update read replicas mode."""
  if args.IsSpecified('read_replicas_mode'):
    patch_request = AddFieldToUpdateMask('read_replicas_mode', patch_request)
  return patch_request


def UpdateSecondaryIpRange(unused_instance_ref, args, patch_request):
  """Hook to update secondary IP range."""
  if args.IsSpecified('secondary_ip_range'):
    patch_request = AddFieldToUpdateMask('secondary_ip_range', patch_request)
  return patch_request


def UpdateMaintenanceVersion(unused_instance_ref, args, patch_request):
  """Hook to update maintenance version to the update mask of the request."""
  if args.IsSpecified('maintenance_version'):
    patch_request = AddFieldToUpdateMask('maintenance_version', patch_request)
  return patch_request



# -*- coding: utf-8 -*- #
# Copyright 2014 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common functions and classes for dealing with managed instances groups."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import json
import random
import re
import string
import sys
from apitools.base.py import encoding
from apitools.base.py import list_pager

from googlecloudsdk.api_lib.compute import exceptions
from googlecloudsdk.api_lib.compute import instance_utils
from googlecloudsdk.api_lib.compute import lister
from googlecloudsdk.api_lib.compute import path_simplifier
from googlecloudsdk.api_lib.compute import request_helper
from googlecloudsdk.api_lib.compute import utils
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as calliope_exceptions
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute.managed_instance_groups import auto_healing_utils
from googlecloudsdk.command_lib.compute.managed_instance_groups import update_instances_utils
from googlecloudsdk.command_lib.util.apis import arg_utils
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core.console import console_io
import six
from six.moves import range  # pylint: disable=redefined-builtin


_ALLOWED_UTILIZATION_TARGET_TYPES = [
    'DELTA_PER_MINUTE', 'DELTA_PER_SECOND', 'GAUGE']

_ALLOWED_UTILIZATION_TARGET_TYPES_LOWER = [
    'delta-per-minute', 'delta-per-second', 'gauge']

ARGS_CONFLICTING_WITH_AUTOSCALING_FILE_BETA = [
    'max_num_replicas', 'cool_down_period', 'custom_metric_utilization',
    'description', 'min_num_replicas',
    'scale_based_on_cpu', 'scale_based_on_load_balancing',
    'target_cpu_utilization', 'target_load_balancing_utilization'
]

ARGS_CONFLICTING_WITH_AUTOSCALING_FILE_ALPHA = (
    ARGS_CONFLICTING_WITH_AUTOSCALING_FILE_BETA)


_MAX_AUTOSCALER_NAME_LENGTH = 63
# 4 character chosen from between lowercase letters and numbers give >1.6M
# possibilities with no more than 100 Autoscalers in one Zone and Project
# so probability that adding an autoscaler will fail because of name conflict
# is about 6e-5.
_NUM_RANDOM_CHARACTERS_IN_AS_NAME = 4


class Error(exceptions.Error):
  """Base exception for managed instance group exceptions."""


class ResourceNotFoundException(Error):
  """The user tries to get/use/update resource which does not exist."""


class InvalidArgumentError(exceptions.Error):
  """The user provides invalid arguments."""


class ResourceAlreadyExistsException(Error):
  """The user tries to create resource which already exists."""


class ResourceCannotBeResolvedException(Error):
  """The user uses invalid / partial name to resolve URI for the resource."""


def AddAutoscalerArgs(
    parser,
    autoscaling_file_enabled=False,
    patch_args=False):
  """Adds commandline arguments to parser."""
  parser.add_argument(
      '--cool-down-period',
      type=arg_parsers.Duration(),
      help=('The time period that the autoscaler should wait before it starts '
            'collecting information from a new instance. This prevents the '
            'autoscaler from collecting information when the instance is '
            'initializing, during which the collected usage would not be '
            'reliable. The default is 60s. See $ gcloud topic datetimes '
            'for information on duration formats.'))
  parser.add_argument('--description', help='Notes about Autoscaler.')
  AddMinMaxControl(parser, max_required=not autoscaling_file_enabled)
  parser.add_argument('--scale-based-on-cpu',
                      action='store_true',
                      help='Autoscaler will be based on CPU utilization.')
  parser.add_argument('--scale-based-on-load-balancing',
                      action='store_true',
                      help=('Use autoscaling based on load balancing '
                            'utilization.'))
  parser.add_argument('--target-cpu-utilization',
                      type=arg_parsers.BoundedFloat(0.0, 1.0),
                      help='Autoscaler aims to maintain CPU utilization at '
                      'target level (0.0 to 1.0).')
  parser.add_argument('--target-load-balancing-utilization',
                      type=arg_parsers.BoundedFloat(0.0, None),
                      help='Autoscaler aims to maintain the load balancing '
                      'utilization level (greater than 0.0).')
  custom_metric_utilization_help = """\
Adds a target metric value for the Autoscaler to use.

*metric*::: Protocol-free URL of a Google Cloud Monitoring metric.

*utilization-target*::: Value of the metric Autoscaler aims to
  maintain (greater than 0.0).

*utilization-target-type*::: How target is expressed. Valid values: {0}.

Mutually exclusive with `--update-stackdriver-metric`.
""".format(', '.join(_ALLOWED_UTILIZATION_TARGET_TYPES))
  parser.add_argument(
      '--custom-metric-utilization',
      type=arg_parsers.ArgDict(
          spec={
              'metric': str,
              'utilization-target': float,
              'utilization-target-type': str,
          },
      ),
      action='append',
      help=custom_metric_utilization_help,
  )

  if autoscaling_file_enabled:
    parser.add_argument(
        '--autoscaling-file',
        metavar='PATH',
        help=('Path of the file from which autoscaling configuration will be '
              'loaded. This flag allows you to atomically setup complex '
              'autoscalers.'))
  parser.add_argument(
      '--remove-stackdriver-metric',
      metavar='METRIC',
      help=('Stackdriver metric to remove from autoscaling configuration. '
            'If the metric is the only input used for autoscaling the '
            'command will fail.'))
  parser.add_argument(
      '--update-stackdriver-metric',
      metavar='METRIC',
      help=('Stackdriver metric to use as an input for autoscaling. '
            'When using this flag, the target value of the metric must also be '
            'specified by using the following flags: '
            '`--stackdriver-metric-single-instance-assignment` or '
            '`--stackdriver-metric-utilization-target` and '
            '`--stackdriver-metric-utilization-target-type`. '
            'Mutually exclusive with `--custom-metric-utilization`.'))
  parser.add_argument(
      '--stackdriver-metric-filter',
      metavar='FILTER',
      help=('Expression for filtering samples used to autoscale, see '
            'https://cloud.google.com/monitoring/api/v3/filters.'))
  parser.add_argument(
      '--stackdriver-metric-utilization-target',
      metavar='TARGET',
      type=float,
      help=('Value of the metric Autoscaler aims to maintain. When '
            'specifying this flag you must also provide '
            '`--stackdriver-metric-utilization-target-type`. Mutually '
            'exclusive with '
            '`--stackdriver-metric-single-instance-assignment` and '
            '`--custom-metric-utilization`.'))

  parser.add_argument(
      '--stackdriver-metric-utilization-target-type',
      metavar='TARGET_TYPE',
      choices=_ALLOWED_UTILIZATION_TARGET_TYPES_LOWER,
      help=('Value of the metric Autoscaler aims to maintain. When '
            'specifying this flag you must also provide '
            '`--stackdriver-metric-utilization-target`. Mutually '
            'exclusive with '
            '`--stackdriver-metric-single-instance-assignment` and '
            '`--custom-metric-utilization`.'))
  parser.add_argument(
      '--stackdriver-metric-single-instance-assignment',
      metavar='ASSIGNMENT',
      type=float,
      help=('Value that indicates the amount of work that each instance is '
            'expected to handle. Autoscaler maintains enough VMs by dividing '
            'the available work by this value. Mutually exclusive with '
            '`-stackdriver-metric-utilization-target-type`, '
            '`-stackdriver-metric-utilization-target-type`, and '
            '`--custom-metric-utilization`.'))

  GetModeFlag().AddToParser(parser)

  AddScaleInControlFlag(parser)

  AddScheduledAutoscaling(parser, patch_args)


def AddMinMaxControl(parser, max_required=True):
  """Adds min and max num replicas controls to a given parser."""
  parser.add_argument('--min-num-replicas',
                      type=arg_parsers.BoundedInt(0, sys.maxsize),
                      help='Minimum number of replicas Autoscaler can set.')
  parser.add_argument('--max-num-replicas',
                      type=arg_parsers.BoundedInt(0, sys.maxsize),
                      required=max_required,
                      help='Maximum number of replicas Autoscaler can set.')


def GetModeFlag():
  # Can't use a ChoiceEnumMapper because we don't have access to the "messages"
  # module for the right API version.
  return base.ChoiceArgument(
      '--mode',
      {
          'on': ('Permits autoscaling to scale out and in (default for '
                 'new autoscalers).'),
          'only-scale-out': ('Permits autoscaling to scale only out and '
                             'not in.'),
          'only-up': ("""
              (DEPRECATED) Permits autoscaling to scale only out and not in.

              Value `only-up` is deprecated. Use `--mode only-scale-out`
              instead.
          """),
          'off': ('Turns off autoscaling, while keeping the new '
                  'configuration.')
      },
      help_str="""\
          Set the mode of an autoscaler for a managed instance group.

          You can turn off or restrict a group's autoscaler activities without
          affecting your autoscaler configuration. The autoscaler configuration
          persists while the activities are turned off or restricted, and the
          activities resume when the autoscaler is turned on again or when the
          restrictions are lifted.
      """)


def AddScaleInControlFlag(parser, include_clear=False):
  """Adds scale-in-control flags to the given parser."""
  arg_group = parser
  # Convert to arg group if --clear-scale-in-control flag is included.
  if include_clear:
    arg_group = parser.add_group(mutex=True)
    arg_group.add_argument(
        '--clear-scale-in-control',
        action='store_true',
        help="""\
          If specified, the scale-in-control field will be cleared. Using this
          flag will remove any configuration set by `--scale-in-control` flag.
        """)
  arg_group.add_argument(
      '--scale-in-control',
      type=arg_parsers.ArgDict(
          spec={
              'max-scaled-in-replicas': str,
              'max-scaled-in-replicas-percent': str,
              'time-window': int,
          },),
      help="""\
        Configuration that allows slower scale in so that even if Autoscaler
        recommends an abrupt scale in of a managed instance group, it will be
        throttled as specified by the parameters.

        *max-scaled-in-replicas*::: Maximum allowed number of VMs that can be
        deducted from the peak recommendation during the window. Possibly all
        these VMs can be deleted at once so the application needs to be prepared
        to lose that many VMs in one step. Mutually exclusive with
        'max-scaled-in-replicas-percent'.

        *max-scaled-in-replicas-percent*::: Maximum allowed percent of VMs
        that can be deducted from the peak recommendation during the window.
        Possibly all these VMs can be deleted at once so the application needs
        to be prepared to lose that many VMs in one step. Mutually exclusive
        with  'max-scaled-in-replicas'.

        *time-window*::: How long back autoscaling should look when computing
        recommendations. The autoscaler will not resize below the maximum
        allowed deduction subtracted from the peak size observed in this
        period. Measured in seconds.
        """)


def AddClearScaleDownControlFlag(parser):
  """Adds --clear-scale-down-control flag to the given parser."""
  arg_group = parser
  arg_group.add_argument(
      '--clear-scale-down-control',
      action='store_true',
      help="""\
        If specified, the scale-down-control field will be cleared. Using this
        flag will remove any configuration set by the now-deprecated
        `--scale-down-control` flag. This is only useful if the MIG
        configuration had scale-down-control set in the past.
      """)


def AddPredictiveAutoscaling(parser, standard=True):
  """Add Predictive autoscaling arguments to the parser."""
  choices = {
      'none': """\
(Default) No predictions are made when calculating the number of VM
instances.
""",
      'optimize-availability': """\
Predictive autoscaling predicts the future values of the
scaling metric and scales the group in advance to ensure that new
VM instances are ready in time to cover the predicted peak.
"""
  }
  if standard:
    choices['standard'] = """
    Standard predictive autoscaling  predicts the future values of
    the scaling metric and then scales the group to ensure that new VM
    instances are ready in time to cover the predicted peak."""
  parser.add_argument(
      '--cpu-utilization-predictive-method',
      choices=choices,
      help="""
      Indicates whether to use a predictive algorithm when scaling based on
      CPU.""",
  )


def AddScheduledAutoscaling(parser, patch_args):
  """Add parameters controlling scheduled autoscaling."""
  if patch_args:
    arg_group = parser.add_group(mutex=True)
    arg_group_config = parser.add_group()
    arg_group.add_argument(
        '--set-schedule',
        metavar='SCHEDULE_NAME',
        help='A unique name for the scaling schedule to be configured.')
    arg_group.add_argument(
        '--update-schedule',
        metavar='SCHEDULE_NAME',
        help='Name of the scaling schedule to be updated.')
    arg_group.add_argument(
        '--remove-schedule',
        metavar='SCHEDULE_NAME',
        help="""\
          Name of the scaling schedule to be removed.

          Be careful with this action as scaling schedule deletion cannot be
          undone.

          You can delete any schedule regardless of its status. If you delete
          a scaling schedule that is currently active, the deleted scaling
          schedule stops being effective immediately after it is deleted.
          If there is no need to maintain capacity, the autoscaler starts
          removing instances after the usual stabilization period and after
          scale-in controls (if configured). For more information, see
          [Delays in scaling in](https://cloud.google.com/compute/docs/autoscaler/understanding-autoscaler-decisions#delays_in_scaling_in) and [Scale-in controls](https://cloud.google.com/compute/docs/autoscaler/understanding-autoscaler-decisions#scale-in_controls).
          This ensures you don't accidentally lose capacity immediately after
          the scaling schedule ends.
          """,
    )
    arg_group.add_argument(
        '--enable-schedule',
        metavar='SCHEDULE_NAME',
        help="""\
        Name of the scaling schedule to be enabled.

        See --disable-schedule for details.
        """)
    arg_group.add_argument(
        '--disable-schedule',
        metavar='SCHEDULE_NAME',
        help="""\
          Name of the scaling schedule to be disabled.

          When a scaling schedule is disabled its configuration persists but
          the scaling schedule itself never becomes active. If you disable a
          scaling schedule that is currently active the disabled scaling
          schedule stops being effective immediately after it moves into
          DISABLED state.
          If there is no need to maintain capacity, the autoscaler starts
          removing instances after the usual stabilization period and after
          scale-in controls (if configured). For more information, see
          [Delays in scaling in](https://cloud.google.com/compute/docs/autoscaler/understanding-autoscaler-decisions#delays_in_scaling_in) and [Scale-in controls](https://cloud.google.com/compute/docs/autoscaler/understanding-autoscaler-decisions#scale-in_controls).
          This ensures you don't accidentally lose capacity immediately after
          the scaling schedule ends.
          """,
    )
    AddScheduledAutoscalingConfigurationArguments(arg_group_config)
  else:
    arg_group = parser
    parser.add_argument(
        '--set-schedule',
        metavar='SCHEDULE_NAME',
        help='Unique name for the scaling schedule.')
    AddScheduledAutoscalingConfigurationArguments(parser.add_group())


def AddScheduledAutoscalingConfigurationArguments(arg_group):
  """Add arguments that are common to adding or modifying a scaling schedule."""
  arg_group.add_argument(
      '--schedule-cron',
      metavar='CRON_EXPRESSION',
      help="""\
        Start time of the scaling schedule in cron format.

        This is when the autoscaler starts creating new VMs, if the group's
        current size is less than the minimum required instances. Set the start
        time to allow enough time for new VMs to boot and initialize. For
        example if your workload takes 10 minutes from VM creation to start
        serving then set the start time 10 minutes earlier than the time you
        need VMs to be ready.
        """)
  arg_group.add_argument(
      '--schedule-duration-sec',
      metavar='DURATION',
      type=arg_parsers.BoundedInt(300, sys.maxsize),
      help="""\
        How long should the scaling schedule be active, measured in seconds.

        Minimum duration is 5 minutes. A scaling schedule is active from its
        start time and for its configured duration. During this time, the
        autoscaler scales the group to have at least as many VMs as defined by
        the minimum required instances. After the configured duration, if there
        is no need to maintain capacity, the autoscaler starts removing
        instances after the usual stabilization period and after scale-in
        controls (if configured). For more information, see
        [Delays in scaling in](https://cloud.google.com/compute/docs/autoscaler/understanding-autoscaler-decisions#delays_in_scaling_in) and [Scale-in controls](https://cloud.google.com/compute/docs/autoscaler/understanding-autoscaler-decisions#scale-in_controls).
        This ensures you don't accidentally lose capacity immediately after
        the scaling schedule ends.
        """,
  )
  arg_group.add_argument(
      '--schedule-min-required-replicas',
      metavar='MIN_REQUIRED_REPLICAS',
      type=arg_parsers.BoundedInt(0, sys.maxsize),
      help="""\
        How many VMs the autoscaler should provision for the duration of this
        scaling schedule.

        Autoscaler provides at least this number of instances when the scaling
        schedule is active. A managed instance group can have more VMs if there
        are other scaling schedules active with more required instances or if
        another signal (for example, scaling based on CPU) requires more
        instances to meet its target.


        This configuration does not change autoscaling minimum and maximum
        instance limits which are always in effect. Autoscaler does not create
        more than the maximum number of instances configured for a group.
        """)
  arg_group.add_argument(
      '--schedule-time-zone',
      metavar='TIME_ZONE',
      help="""\
        Name of the timezone that the scaling schedule's start time is in.

        It should be provided as a name from the IANA tz database (for
        example Europe/Paris or UTC). It automatically adjusts for daylight
        savings time (DST). If no time zone is provided, UTC is used as a
        default.

        See https://en.wikipedia.org/wiki/List_of_tz_database_time_zones for
        the list of valid timezones.
        """)
  arg_group.add_argument(
      '--schedule-description',
      metavar='DESCRIPTION',
      help='A verbose description of the scaling schedule.')


def ValidateConflictsWithAutoscalingFile(args, conflicting_args):
  if (hasattr(args, 'autoscaling_file') and
      args.IsSpecified('autoscaling_file')):
    for arg in conflicting_args:
      if args.IsSpecified(arg):
        conflicting_flags = [
            '--' + a.replace('_', '-')
            for a in conflicting_args
        ]
        raise calliope_exceptions.ConflictingArgumentsException(
            *(['--autoscaling-file'] + conflicting_flags))


def _ValidateCustomMetricUtilizationVsUpdateStackdriverMetric(args):
  if (args.IsSpecified('custom_metric_utilization') and
      args.IsSpecified('update_stackdriver_metric')):
    raise calliope_exceptions.ConflictingArgumentsException(
        '--custom-metric-utilization', '--update-stackdriver-metric')


def _ValidateRemoveStackdriverMetricVsUpdateStackdriverMetric(args):
  if (args.IsSpecified('update_stackdriver_metric') and
      args.IsSpecified('remove_stackdriver_metric') and
      args.update_stackdriver_metric == args.remove_stackdriver_metric):
    raise calliope_exceptions.InvalidArgumentException(
        '--update-stackdriver-metric',
        'You can not remove Stackdriver metric you are updating with '
        '[--update-stackdriver-metric] flag.')


def _ValidateRequiringUpdateStackdriverMetric(args):  # pylint:disable=missing-docstring
  if not args.IsSpecified('update_stackdriver_metric'):
    requiring_flags = [
        'stackdriver_metric_filter',
        'stackdriver_metric_single_instance_assignment',
        'stackdriver_metric_utilization_target',
        'stackdriver_metric_utilization_target_type',
    ]
    for f in requiring_flags:
      if args.IsSpecified(f):
        raise calliope_exceptions.RequiredArgumentException(
            '--' + f.replace('_', '-'),
            '[--update-stackdriver-metric] required to use this flag.')


def _ValidateRequiredByUpdateStackdriverMetric(args):  # pylint:disable=missing-docstring
  if args.IsSpecified('update_stackdriver_metric'):
    one_of_required = [
        'stackdriver_metric_single_instance_assignment',
        'stackdriver_metric_utilization_target',]
    if not any([args.IsSpecified(f) for f in one_of_required]):
      flags = [
          '[--{}]'.format(f.replace('_', '--')) for f in one_of_required]
      msg = ('You must provide one of {} with '
             '[--update-stackdriver-metric].'.format(', '.join(flags)))
      raise calliope_exceptions.RequiredArgumentException(
          '--update-stackdriver-metric', msg)


def _ValidateSingleInstanceAssignmentVsUtilizationTarget(args):  # pylint:disable=missing-docstring
  if args.IsSpecified('stackdriver_metric_single_instance_assignment'):
    potential_conflicting = [
        'stackdriver_metric_utilization_target',
        'stackdriver_metric_utilization_target_type',
    ]
    conflicting = [f for f in potential_conflicting if args.IsSpecified(f)]
    if any(conflicting):
      assignment_flag = '--stackdriver-metric-single-instance-assignment'
      conflicting_flags = [
          '[--{}]'.format(f.replace('_', '-')) for f in conflicting]
      raise calliope_exceptions.ConflictingArgumentsException(
          assignment_flag,
          'You cannot use any of {} with `{}`'.format(
              conflicting_flags, assignment_flag))


def _ValidateUtilizationTargetHasType(args):
  if (args.IsSpecified('stackdriver_metric_utilization_target') and
      not args.IsSpecified('stackdriver_metric_utilization_target_type')):
    raise calliope_exceptions.RequiredArgumentException(
        '--stackdriver-metric-utilization-target-type',
        'Required with [--stackdriver-metric-utilization-target].')


def ValidateStackdriverMetricsFlags(args):
  """Perform validations related to .*stackdriver-metric.* flags."""
  _ValidateCustomMetricUtilizationVsUpdateStackdriverMetric(args)
  _ValidateRemoveStackdriverMetricVsUpdateStackdriverMetric(args)
  _ValidateRequiringUpdateStackdriverMetric(args)
  _ValidateRequiredByUpdateStackdriverMetric(args)
  _ValidateSingleInstanceAssignmentVsUtilizationTarget(args)
  _ValidateUtilizationTargetHasType(args)


def ValidateGeneratedAutoscalerIsValid(args, autoscaler):
  if (args.IsSpecified('remove_stackdriver_metric') and
      not autoscaler.autoscalingPolicy.customMetricUtilizations and
      not autoscaler.autoscalingPolicy.cpuUtilization and
      not autoscaler.autoscalingPolicy.loadBalancingUtilization):
    raise calliope_exceptions.InvalidArgumentException(
        '--remove-stackdriver-metric',
        'This would remove the only signal used for autoscaling. If you want '
        'to stop autoscaling the Managed Instance Group use `stop-autoscaling` '
        'command instead.')


def ValidateAutoscalerArgs(args):
  """Validates args."""
  if args.min_num_replicas and args.max_num_replicas:
    if args.min_num_replicas > args.max_num_replicas:
      raise calliope_exceptions.InvalidArgumentException(
          '--max-num-replicas', 'can\'t be less than min num replicas.')

  if args.custom_metric_utilization:
    for custom_metric_utilization in args.custom_metric_utilization:
      for field in ('utilization-target', 'metric', 'utilization-target-type'):
        if field not in custom_metric_utilization:
          raise calliope_exceptions.InvalidArgumentException(
              '--custom-metric-utilization', field + ' not present.')
      if custom_metric_utilization['utilization-target'] < 0:
        raise calliope_exceptions.InvalidArgumentException(
            '--custom-metric-utilization utilization-target', 'less than 0.')


def GetInstanceGroupManagerOrThrow(igm_ref, client):
  """Retrieves the given Instance Group Manager if possible.

  Args:
    igm_ref: reference to the Instance Group Manager.
    client: The compute client.
  Returns:
    Instance Group Manager object.
  """
  if hasattr(igm_ref, 'region'):
    service = client.apitools_client.regionInstanceGroupManagers
    request_type = service.GetRequestType('Get')
  if hasattr(igm_ref, 'zone'):
    service = client.apitools_client.instanceGroupManagers
    request_type = service.GetRequestType('Get')
  request = request_type(**igm_ref.AsDict())

  errors = []
  # Run through the generator to actually make the requests and get potential
  # errors.
  igm_details = client.MakeRequests([(service, 'Get', request)],
                                    errors_to_collect=errors)

  if errors or len(igm_details) != 1:
    utils.RaiseException(errors, ResourceNotFoundException,
                         error_message='Could not fetch resource:')
  return igm_details[0]


def CreateZoneRef(resources, data):
  """Create zone reference from object with project and zone fields."""
  return resources.Parse(
      None,
      params={'project': data.project,
              'zone': data.zone},
      collection='compute.zones')


def CreateRegionRef(resources, data):
  """Create region reference from object with project and region fields."""
  return resources.Parse(
      None,
      params={'project': data.project,
              'region': data.region},
      collection='compute.regions')


def GroupByProject(locations):
  """Group locations by project field."""
  result = {}
  for location in locations or []:
    if location.project not in result:
      result[location.project] = []
    result[location.project].append(location)
  return result


def AutoscalersForLocations(zones, regions, client,
                            fail_when_api_not_supported=True):
  """Finds all Autoscalers defined for a given project and locations.

  Args:
    zones: iterable of target zone references
    regions: iterable of target region references
    client: The compute client.
    fail_when_api_not_supported: If true, raise tool exception if API does not
        support autoscaling.
  Returns:
    A list of Autoscaler objects.
  """
  # Errors is passed through library calls and modified with
  # (ERROR_CODE, ERROR_MESSAGE) tuples.
  errors = []

  # Explicit list() is required to unwind the generator and make sure errors
  # are detected at this level.
  requests = []
  for project, zones in six.iteritems(GroupByProject(zones)):
    requests += lister.FormatListRequests(
        service=client.apitools_client.autoscalers,
        project=project,
        scopes=sorted(set([zone_ref.zone for zone_ref in zones])),
        scope_name='zone',
        filter_expr=None)

  if regions:
    if hasattr(client.apitools_client, 'regionAutoscalers'):
      for project, regions in six.iteritems(GroupByProject(regions)):
        requests += lister.FormatListRequests(
            service=client.apitools_client.regionAutoscalers,
            project=project,
            scopes=sorted(set([region_ref.region for region_ref in regions])),
            scope_name='region',
            filter_expr=None)
    else:
      if fail_when_api_not_supported:
        errors.append((None, 'API does not support regional autoscaling'))

  autoscalers = client.MakeRequests(
      requests=requests,
      errors_to_collect=errors)

  if errors:
    utils.RaiseToolException(
        errors,
        error_message='Could not check if the Managed Instance Group is '
        'Autoscaled.')

  return autoscalers


def AutoscalersForMigs(migs, autoscalers):
  """Finds Autoscalers with target amongst given IGMs.

  Args:
    migs: List of triples (IGM name, scope type, location reference).
    autoscalers: A list of Autoscalers to search among.
  Returns:
    A list of all Autoscalers with target on mig_names list.
  """
  igm_url_regexes = []
  for (name, scope_type, location) in migs:
    igm_url_regexes.append(
        '/projects/{project}/{scopeType}/{scopeName}/'
        'instanceGroupManagers/{name}$'
        .format(project=location.project,
                scopeType=(scope_type + 's'),
                scopeName=getattr(location, scope_type),
                name=name))
  igm_url_regex = re.compile('(' + ')|('.join(igm_url_regexes) + ')')
  result = [
      autoscaler for autoscaler in autoscalers
      if igm_url_regex.search(autoscaler.target)
  ]
  return result


def AutoscalerForMigByRef(client, resources, igm_ref):
  """Returns autoscaler targeting given instance group manager.

  Args:
    client: a GCE client
    resources: a GCE resource registry
    igm_ref: reference to instance group manager
  Returns:
    Autoscaler message with autoscaler targeting the IGM refferenced by
    igm_ref or None if there isn't one.
  Raises:
    ValueError: if instance group manager collection path is unknown
  """
  if igm_ref.Collection() == 'compute.instanceGroupManagers':
    scope_type = 'zone'
    location = CreateZoneRef(resources, igm_ref)
    zones, regions = [location], None
  elif igm_ref.Collection() == 'compute.regionInstanceGroupManagers':
    scope_type = 'region'
    location = CreateRegionRef(resources, igm_ref)
    zones, regions = None, [location]
  else:
    raise ValueError('Unknown reference type {0}'.format(igm_ref.Collection()))

  autoscalers = AutoscalersForLocations(
      regions=regions,
      zones=zones,
      client=client)

  return AutoscalerForMig(
      mig_name=igm_ref.Name(),
      autoscalers=autoscalers,
      location=location,
      scope_type=scope_type)


def AutoscalerForMig(mig_name, autoscalers, location, scope_type):
  """Finds Autoscaler targeting given IGM.

  Args:
    mig_name: Name of MIG targeted by Autoscaler.
    autoscalers: A list of Autoscalers to search among.
    location: Target location reference.
    scope_type: Target scope type.
  Returns:
    Autoscaler object for autoscaling the given Instance Group Manager or None
    when such Autoscaler does not exist.
  Raises:
    InvalidArgumentError: if more than one autoscaler provided
  """
  autoscalers = AutoscalersForMigs(
      [(mig_name, scope_type, location)], autoscalers)
  if autoscalers:
    # For each Instance Group Manager there can be at most one Autoscaler having
    # the Manager as a target, so when one is found it can be returned as it is
    # the only one.
    if len(autoscalers) == 1:
      return autoscalers[0]
    else:
      raise InvalidArgumentError(
          'More than one Autoscaler with given target.')
  return None


def AddAutoscalersToMigs(migs_iterator,
                         client,
                         resources,
                         fail_when_api_not_supported=True):
  """Add Autoscaler to each IGM object if autoscaling is enabled for it."""
  def ParseZone(zone_link):
    return resources.Parse(
        zone_link,
        params={'project': properties.VALUES.core.project.GetOrFail},
        collection='compute.zones')

  def ParseRegion(region_link):
    return resources.Parse(
        region_link,
        params={'project': properties.VALUES.core.project.GetOrFail},
        collection='compute.regions')

  migs = list(migs_iterator)
  zones = set([ParseZone(mig['zone']) for mig in migs if 'zone' in mig])
  regions = set(
      [ParseRegion(mig['region']) for mig in migs if 'region' in mig])
  autoscalers = {}
  all_autoscalers = AutoscalersForLocations(
      zones=zones,
      regions=regions,
      client=client,
      fail_when_api_not_supported=fail_when_api_not_supported)

  for location in list(zones) + list(regions):
    autoscalers[location.Name()] = []

  for autoscaler in all_autoscalers:
    autoscaler_scope = None
    if autoscaler.zone is not None:
      autoscaler_scope = ParseZone(autoscaler.zone)
    if hasattr(autoscaler, 'region') and autoscaler.region is not None:
      autoscaler_scope = ParseRegion(autoscaler.region)
    if autoscaler_scope is not None:
      autoscalers.setdefault(autoscaler_scope.Name(), [])
      autoscalers[autoscaler_scope.Name()].append(autoscaler)

  for mig in migs:
    location = None
    scope_type = None
    if 'region' in mig:
      location = ParseRegion(mig['region'])
      scope_type = 'region'
    elif 'zone' in mig:
      location = ParseZone(mig['zone'])
      scope_type = 'zone'

    autoscaler = None
    if location and scope_type:
      autoscaler = AutoscalerForMig(
          mig_name=mig['name'],
          autoscalers=autoscalers[location.Name()],
          location=location,
          scope_type=scope_type)
    if autoscaler:
      mig['autoscaler'] = autoscaler
    yield mig


def _BuildCpuUtilization(args, messages):
  """Builds the CPU Utilization message given relevant arguments."""
  flags_to_check = ['target_cpu_utilization', 'scale_based_on_cpu',
                    'cpu_utilization_predictive_method']

  if instance_utils.IsAnySpecified(args, *flags_to_check):
    cpu_message = messages.AutoscalingPolicyCpuUtilization()
    if args.target_cpu_utilization:
      cpu_message.utilizationTarget = args.target_cpu_utilization
    if args.cpu_utilization_predictive_method:
      cpu_predictive_enum = messages.AutoscalingPolicyCpuUtilization.PredictiveMethodValueValuesEnum
      cpu_message.predictiveMethod = arg_utils.ChoiceToEnum(
          args.cpu_utilization_predictive_method, cpu_predictive_enum)
    return cpu_message
  return None


def _BuildCustomMetricUtilizationsFromCustomMetricUtilizationFlag(
    flag, messages):
  """Translate --custom-metric-utilization flag to API message."""
  result = []
  for custom_metric_utilization in flag:
    result.append(
        messages.AutoscalingPolicyCustomMetricUtilization(
            utilizationTarget=custom_metric_utilization[
                'utilization-target'],
            metric=custom_metric_utilization['metric'],
            utilizationTargetType=(
                messages
                .AutoscalingPolicyCustomMetricUtilization
                .UtilizationTargetTypeValueValuesEnum(
                    custom_metric_utilization['utilization-target-type'],
                )
            ),
        )
    )
  return result


def _RemoveMetricFromList(metrics, to_remove):
  for i, metric in enumerate(metrics):
    if metric.metric == to_remove:
      del metrics[i]
      return


def _UpdateCustomMetricUtilizationsFromStackoverflowFlags(
    args, messages, original):
  """Take apply stackdriver flags to customMetricUtilizations."""
  if original:
    result = original.autoscalingPolicy.customMetricUtilizations
  else:
    result = []
  if args.remove_stackdriver_metric:
    _RemoveMetricFromList(result, args.remove_stackdriver_metric)
  if args.update_stackdriver_metric:
    _RemoveMetricFromList(result, args.update_stackdriver_metric)
    if args.stackdriver_metric_utilization_target_type:
      target_type = (
          messages.AutoscalingPolicyCustomMetricUtilization
          .UtilizationTargetTypeValueValuesEnum(
              args.stackdriver_metric_utilization_target_type.upper().
              replace('-', '_'),
          )
      )
    else:
      target_type = None

    if args.stackdriver_metric_filter and "'" in args.stackdriver_metric_filter:
      log.warning(
          "The provided filter contains a single quote character ('). While "
          "valid as a metric/resource label value, it's not a control "
          "character that is part of the filtering language; if you meant "
          "to use it to quote a string value, you need to use a double quote "
          "character (\") instead.")
    result.append(
        messages.AutoscalingPolicyCustomMetricUtilization(
            utilizationTarget=args.stackdriver_metric_utilization_target,
            metric=args.update_stackdriver_metric,
            utilizationTargetType=target_type,
            singleInstanceAssignment=(
                args.stackdriver_metric_single_instance_assignment
            ),
            filter=args.stackdriver_metric_filter,
        )
    )
  return result


def _BuildCustomMetricUtilizations(args, messages, original):
  """Builds custom metric utilization policy list from args.

  Args:
    args: command line arguments.
    messages: module containing message classes.
    original: original autoscaler message.
  Returns:
    AutoscalingPolicyCustomMetricUtilization list.
  """
  if args.custom_metric_utilization:
    return _BuildCustomMetricUtilizationsFromCustomMetricUtilizationFlag(
        args.custom_metric_utilization, messages)
  if hasattr(args, 'stackdriver_metric_filter'):
    return _UpdateCustomMetricUtilizationsFromStackoverflowFlags(
        args, messages, original)


def _BuildLoadBalancingUtilization(args, messages):
  if args.target_load_balancing_utilization:
    return messages.AutoscalingPolicyLoadBalancingUtilization(
        utilizationTarget=args.target_load_balancing_utilization,
    )
  if args.scale_based_on_load_balancing:
    return messages.AutoscalingPolicyLoadBalancingUtilization()
  return None


def ParseModeString(mode, messages):
  return messages.AutoscalingPolicy.ModeValueValuesEnum(
      mode.upper().replace('-', '_'))


def _BuildMode(args, messages, original):
  if not args.mode:
    return original.autoscalingPolicy.mode if original else None
  return ParseModeString(args.mode, messages)


def BuildScaleIn(args, messages):
  """Builds AutoscalingPolicyScaleInControl.

  Args:
    args: command line arguments.
    messages: module containing message classes.
  Returns:
    AutoscalingPolicyScaleInControl message object.
  Raises:
    InvalidArgumentError:  if both max-scaled-in-replicas and
      max-scaled-in-replicas-percent are specified.
  """
  if args.IsSpecified('scale_in_control'):
    replicas_arg = args.scale_in_control.get('max-scaled-in-replicas')
    replicas_arg_percent = args.scale_in_control.get(
        'max-scaled-in-replicas-percent')
    if replicas_arg and replicas_arg_percent:
      raise InvalidArgumentError(
          'max-scaled-in-replicas and max-scaled-in-replicas-percent'
          'are mutually exclusive, you can\'t specify both')
    elif replicas_arg_percent:
      max_replicas = messages.FixedOrPercent(percent=int(replicas_arg_percent))
    else:
      max_replicas = messages.FixedOrPercent(fixed=int(replicas_arg))

    return messages.AutoscalingPolicyScaleInControl(
        maxScaledInReplicas=max_replicas,
        timeWindowSec=args.scale_in_control.get('time-window'))


def _RemoveScheduleEncoder(message, unused_encoder=None):
  """Encoder for use when removing a schedule.

  It works around issues with proto encoding of AdditionalProperties with null
  values by directly encoding a dict of keys with None values into json,
  skipping proto-based encoding.

  Args:
    message: an instance of AutoscalingPolicy.ScalingSchedulesValue.
  Returns:
    Schedule removal request JSON dumped to string.
  """
  py_object = {property.key: None for property in message.additionalProperties}
  return json.dumps(py_object)


def _RemoveScheduleDecoder(unused_data, unused_decoder=None):
  """Placeholder decoder for schedule removal message.

  It's passed when registering message codec, but it will never be used as
  removing schedules is a write-only operation.
  """
  raise NotImplementedError("This should never be called, it's a write-only"
                            " operation.")


def BuildSchedules(args, messages):
  """Builds AutoscalingPolicyScalingSchedules.

  Args:
    args: command line arguments.
    messages: module containing message classes.
  Returns:
    Dict containing an AutoscalingPolicyScalingSchedule message object.
  Raises:
    InvalidArgumentError:  if more than one of --scaling-schedule,
    --update-schedule, --remove-schedule,
    --enable-schedule, --disable-schedule is specified.
  """
  # The following method uses getattr instead of args.IsSpecified because the
  # latter doesn't work before the parameter is available in GA, raising
  # googlecloudsdk.calliope.parser_errors.UnknownDestinationException.
  mutex_group = {'set_schedule', 'update_schedule', 'remove_schedule',
                 'enable_schedule', 'disable_schedule'}
  count = 0
  for possible_argument in mutex_group:
    if getattr(args, possible_argument, None) is not None:
      count += 1
  if count == 0:
    return None
  if count > 1:
    raise InvalidArgumentError(
        '--set-schedule, --update-schedule, --remove-schedule, '
        '--enable-schedule, --disable-schedule are mutually exclusive, only '
        'one can be specified.')

  scaling_schedule_wrapper = (messages.AutoscalingPolicy.
                              ScalingSchedulesValue.AdditionalProperty)
  field_mapping = {
      'schedule_cron': 'schedule',
      'schedule_duration_sec': 'durationSec',
      'schedule_min_required_replicas': 'minRequiredReplicas',
      'schedule_time_zone': 'timeZone',
      'schedule_description': 'description',
  }

  if getattr(args, 'enable_schedule', None) is not None:
    return messages.AutoscalingPolicy.ScalingSchedulesValue(
        additionalProperties=[
            scaling_schedule_wrapper(
                key=args.enable_schedule,
                value=messages.AutoscalingPolicyScalingSchedule(
                    disabled=False))])
  if getattr(args, 'disable_schedule', None) is not None:
    return messages.AutoscalingPolicy.ScalingSchedulesValue(
        additionalProperties=[
            scaling_schedule_wrapper(
                key=args.disable_schedule,
                value=messages.AutoscalingPolicyScalingSchedule(
                    disabled=True))])
  if getattr(args, 'remove_schedule', None) is not None:
    # Register a custom message codec to properly encode AdditionalProperty's
    # None value into nulls.
    encoding.RegisterCustomMessageCodec(
        encoder=_RemoveScheduleEncoder,
        decoder=_RemoveScheduleDecoder)(
            messages.AutoscalingPolicy.ScalingSchedulesValue)
    return messages.AutoscalingPolicy.ScalingSchedulesValue(
        additionalProperties=[
            scaling_schedule_wrapper(
                key=args.remove_schedule,
                value=None)])
  if getattr(args, 'set_schedule', None) is not None:
    policy_name = args.set_schedule
    required = {'schedule_cron', 'schedule_duration_sec',
                'schedule_min_required_replicas'}
    # Set-shedule should clear pre-existing fields, so we set them to None.
    scaling_schedule = {field: None for field in field_mapping.values()}
  else:
    policy_name = args.update_schedule
    required = set()
    scaling_schedule = {}

  for arg_attr, field in six.iteritems(field_mapping):
    arg = getattr(args, arg_attr, None)
    if arg is not None:
      scaling_schedule[field] = arg
    elif arg_attr in required:
      raise InvalidArgumentError(
          '--set-schedule argument requires --schedule-duration-sec, '
          '--schedule-cron, and --schedule-min-required-replicas to be '
          'specified.')

  return messages.AutoscalingPolicy.ScalingSchedulesValue(
      additionalProperties=[
          scaling_schedule_wrapper(
              key=policy_name,
              value=messages.AutoscalingPolicyScalingSchedule(
                  **scaling_schedule))])


def _BuildAutoscalerPolicy(args, messages, original):
  """Builds AutoscalingPolicy from args.

  Args:
    args: command line arguments.
    messages: module containing message classes.
    original: original autoscaler message.
  Returns:
    AutoscalingPolicy message object.
  """
  policy_dict = {
      'coolDownPeriodSec': args.cool_down_period,
      'cpuUtilization': _BuildCpuUtilization(args, messages),
      'customMetricUtilizations': _BuildCustomMetricUtilizations(
          args, messages, original),
      'loadBalancingUtilization': _BuildLoadBalancingUtilization(
          args, messages),
      'maxNumReplicas': args.max_num_replicas,
      'minNumReplicas': args.min_num_replicas,
  }
  policy_dict['mode'] = _BuildMode(args, messages, original)
  policy_dict['scaleInControl'] = BuildScaleIn(args, messages)
  policy_dict['scalingSchedules'] = BuildSchedules(args, messages)

  return messages.AutoscalingPolicy(
      **dict((key, value) for key, value in six.iteritems(policy_dict)
             if value is not None))  # Filter out None values.


def AdjustAutoscalerNameForCreation(autoscaler_resource, igm_ref):
  """Set name of autoscaler o be created.

  If autoscaler name is not None it wNone ill be used as a prefix of name of the
  autoscaler to be created. Prefix may be shortened so that the name fits below
  length limit. Name prefix is followed by '-' character and four
  random letters.

  Args:
    autoscaler_resource: Autoscaler resource to be created.
    igm_ref: reference to Instance Group Manager targeted by the Autoscaler.
  """
  if autoscaler_resource.name is None:
    autoscaler_resource.name = igm_ref.Name()
  trimmed_name = autoscaler_resource.name[
      0:(_MAX_AUTOSCALER_NAME_LENGTH - _NUM_RANDOM_CHARACTERS_IN_AS_NAME - 1)]
  random_characters = [
      random.choice(string.ascii_lowercase + string.digits)
      for _ in range(_NUM_RANDOM_CHARACTERS_IN_AS_NAME)
  ]
  random_suffix = ''.join(random_characters)
  new_name = '{0}-{1}'.format(trimmed_name, random_suffix)
  autoscaler_resource.name = new_name


def BuildAutoscaler(args,
                    messages,
                    igm_ref,
                    name,
                    original):
  """Builds autoscaler message protocol buffer."""
  autoscaler = messages.Autoscaler(
      autoscalingPolicy=_BuildAutoscalerPolicy(
          args,
          messages,
          original),
      description=args.description,
      name=name,
      target=igm_ref.SelfLink(),
  )
  return autoscaler


def CreateAutohealingPolicies(messages, health_check, initial_delay):
  """Creates autohealing policy list from args."""
  if health_check is None and initial_delay is None:
    return []
  policy = messages.InstanceGroupManagerAutoHealingPolicy()
  if health_check:
    policy.healthCheck = health_check
  if initial_delay:
    policy.initialDelaySec = initial_delay
  return [policy]


def ModifyAutohealingPolicies(current_policies, messages, args,
                              health_check_url):
  """Modifies existing autohealing policy from args."""
  if args.clear_autohealing:
    return [messages.InstanceGroupManagerAutoHealingPolicy()]
  if not health_check_url and not args.initial_delay:
    # No modifications of AutoHealingPolicies.
    return None
  current_policy = current_policies[0] if current_policies else None
  policy = current_policy or messages.InstanceGroupManagerAutoHealingPolicy()
  if health_check_url:
    policy.healthCheck = health_check_url
  if args.initial_delay:
    policy.initialDelaySec = args.initial_delay
  return [policy]


def ValidateAutohealingPolicies(auto_healing_policies):
  """Validates autohealing policies.

  Args:
    auto_healing_policies: list of AutoHealingPolicies to validate
  """
  if not auto_healing_policies:
    return
  # Only a single auto_healing_policy is allowed. Displaying warnings for any
  # additional entries is unnecessary as an error will be returned.
  policy = auto_healing_policies[0]
  if not policy.healthCheck and policy.initialDelaySec:
    message = ('WARNING: Health check should be provided when specifying '
               'initial delay.')
    console_io.PromptContinue(message=message, cancel_on_no=True)


def CreateInstanceLifecyclePolicy(messages, args):
  """Creates instance lifecycle policy list from args."""
  policy = messages.InstanceGroupManagerInstanceLifecyclePolicy()
  if args.IsSpecified('force_update_on_repair'):
    if args.force_update_on_repair:
      policy.forceUpdateOnRepair = messages.InstanceGroupManagerInstanceLifecyclePolicy.ForceUpdateOnRepairValueValuesEnum.YES
    else:
      policy.forceUpdateOnRepair = messages.InstanceGroupManagerInstanceLifecyclePolicy.ForceUpdateOnRepairValueValuesEnum.NO

  return ValueOrNone(policy)


def CreateStandbyPolicy(messages, initial_delay_sec):
  """Creates standby policy from args."""
  # pylint: disable=g-explicit-bool-comparison
  if initial_delay_sec is None:
    return None
  return messages.InstanceGroupManagerStandbyPolicy(
      initialDelaySec=initial_delay_sec)


def _GetInstanceTemplatesSet(*versions_lists):
  versions_set = set()
  for versions_list in versions_lists:
    versions_set.update(versions_list)
  return versions_set


def ValidateVersions(igm_info, new_versions, resources, force=False):
  """Validates whether versions provided by user are consistent.

  Args:
    igm_info: instance group manager resource.
    new_versions: list of new versions.
    force: if true, we allow any combination of instance templates, as long as
    they are different. If false, only the following transitions are allowed: X
      -> Y, X -> (X, Y), (X, Y) -> X, (X, Y) -> Y, (X, Y) -> (X, Y)
  Raises:
     InvalidArgumentError: if provided arguments are not complete or invalid.
  """
  if (len(new_versions) == 2 and
      new_versions[0].instanceTemplate == new_versions[1].instanceTemplate):
    raise InvalidArgumentError(
        'Provided instance templates must be different.')
  if force:
    return

  # Only X -> Y, X -> (X, Y), (X, Y) -> X, (X, Y) -> Y, (X, Y) -> (X, Y)
  # are allowed in gcloud (unless --force)
  # Equivalently, at most two versions in old and new versions set union
  if igm_info.versions:
    igm_templates = [
        resources.ParseURL(version.instanceTemplate).RelativeName()
        for version in igm_info.versions
    ]
  elif igm_info.instanceTemplate:
    igm_templates = [
        resources.ParseURL(igm_info.instanceTemplate).RelativeName()
    ]
  else:
    raise InvalidArgumentError(
        'Either versions or instance template must be specified for '
        'managed instance group.')

  new_templates = [
      resources.ParseURL(version.instanceTemplate).RelativeName()
      for version in new_versions
  ]

  version_count = len(_GetInstanceTemplatesSet(igm_templates, new_templates))
  if version_count > 2:
    raise InvalidArgumentError(
        'Update inconsistent with current state. '
        'The only allowed transitions between versions are: '
        'X -> Y, X -> (X, Y), (X, Y) -> X, (X, Y) -> Y, (X, Y) -> (X, Y). '
        'Please check versions templates or use --force.')


def AddAutoscaledPropertyToMigs(migs, client, resources):
  """Add Autoscaler information if Autoscaler is defined for the MIGs.

  Issue additional queries to detect if any given Instange Group Manager is
  a target of some autoscaler and add this information to in 'autoscaled'
  property.

  Args:
    migs: list of dicts, List of IGM resources converted to dictionaries
    client: a GCE client
    resources: a GCE resource registry

  Returns:
    Pair of:
    - boolean - True iff any autoscaler has an error
    - Copy of migs list with additional property 'autoscaled' set to 'No'/'Yes'/
    'Yes (*)' for each MIG depending on look-up result.
  """

  augmented_migs = []
  had_errors = False
  for mig in AddAutoscalersToMigs(
      migs_iterator=_ComputeInstanceGroupSize(migs, client, resources),
      client=client,
      resources=resources,
      fail_when_api_not_supported=False):
    status = ResolveAutoscalingStatusForMig(client, mig)
    if status == client.messages.Autoscaler.StatusValueValuesEnum.ERROR:
      had_errors = True
    augmented_migs.append(mig)
  return (had_errors, augmented_migs)


def ResolveAutoscalingStatusForMig(client, mig):
  """Resolves 'autoscaled' property for MIG.

  Uses 'autoscaler' property of a MIG to resolve 'autoscaled' property for
  output.

  Args:
    client: a GCE client
    mig: IGM resource as a dict

  Returns:
    Status of autoscaler if MIG is autoscaled. None otherwise.
  """
  if 'autoscaler' in mig and mig['autoscaler'] is not None:
    # status is present in autoscaler iff Autoscaler message has embedded
    # StatusValueValuesEnum defined.
    if (hasattr(mig['autoscaler'], 'status') and mig['autoscaler'].status
        == client.messages.Autoscaler.StatusValueValuesEnum.ERROR):
      mig['autoscaled'] = 'yes (*)'
      return mig['autoscaler'].status
    else:
      # Assume it to be ACTIVE
      mig['autoscaled'] = 'yes'
      return client.messages.Autoscaler.StatusValueValuesEnum.ACTIVE
  else:
    mig['autoscaled'] = 'no'
    return None


def _ComputeInstanceGroupSize(items, client, resources):
  """Add information about Instance Group size."""
  errors = []
  zone_refs = [
      resources.Parse(
          mig['zone'],
          params={'project': properties.VALUES.core.project.GetOrFail},
          collection='compute.zones') for mig in items if 'zone' in mig
  ]
  region_refs = [
      resources.Parse(
          mig['region'],
          params={'project': properties.VALUES.core.project.GetOrFail},
          collection='compute.regions') for mig in items if 'region' in mig
  ]

  zonal_instance_groups = []
  for project, zone_refs in six.iteritems(GroupByProject(zone_refs)):
    zonal_instance_groups.extend(
        lister.GetZonalResources(
            service=client.apitools_client.instanceGroups,
            project=project,
            requested_zones=set([zone.zone for zone in zone_refs]),
            filter_expr=None,
            http=client.apitools_client.http,
            batch_url=client.batch_url,
            errors=errors))

  regional_instance_groups = []
  if getattr(client.apitools_client, 'regionInstanceGroups', None):
    for project, region_refs in six.iteritems(GroupByProject(region_refs)):
      regional_instance_groups.extend(
          lister.GetRegionalResources(
              service=client.apitools_client.regionInstanceGroups,
              project=project,
              requested_regions=set([region.region for region in region_refs]),
              filter_expr=None,
              http=client.apitools_client.http,
              batch_url=client.batch_url,
              errors=errors))

  instance_groups = zonal_instance_groups + regional_instance_groups
  instance_group_uri_to_size = {ig.selfLink: ig.size for ig in instance_groups}

  if errors:
    utils.RaiseToolException(errors)

  for item in items:
    self_link = item['selfLink']
    gm_self_link = self_link.replace('/instanceGroupManagers/',
                                     '/instanceGroups/')

    item['size'] = str(instance_group_uri_to_size.get(gm_self_link, ''))
    yield item


def GetHealthCheckUri(resources, args):
  """Creates health check reference from args."""
  if args.health_check:
    ref = auto_healing_utils.HEALTH_CHECK_ARG.ResolveAsResource(
        args,
        resources,
        default_scope=compute_flags.compute_scope.ScopeEnum.GLOBAL)
    return ref.SelfLink()
  if args.http_health_check:
    return resources.Parse(
        args.http_health_check,
        params={'project': properties.VALUES.core.project.GetOrFail},
        collection='compute.httpHealthChecks').SelfLink()
  if args.https_health_check:
    return resources.Parse(
        args.https_health_check,
        params={'project': properties.VALUES.core.project.GetOrFail},
        collection='compute.httpsHealthChecks').SelfLink()


# TODO(b/70203649): improve/fix method (no silent errors, add optimizations)
def CreateInstanceReferences(holder, igm_ref, instance_names):
  """Creates references to instances in instance group (zonal or regional)."""
  if igm_ref.Collection() == 'compute.instanceGroupManagers':
    instance_refs = []
    for instance in instance_names:
      instance_refs.append(
          holder.resources.Parse(
              instance,
              params={
                  'project': igm_ref.project,
                  'zone': igm_ref.zone,
              },
              collection='compute.instances'))
    return instance_refs
  elif igm_ref.Collection() == 'compute.regionInstanceGroupManagers':
    messages = holder.client.messages
    request = (
        messages.ComputeRegionInstanceGroupManagersListManagedInstancesRequest)(
            instanceGroupManager=igm_ref.Name(),
            region=igm_ref.region,
            project=igm_ref.project)
    managed_instances = list_pager.YieldFromList(
        service=holder.client.apitools_client.regionInstanceGroupManagers,
        batch_size=500,
        request=request,
        method='ListManagedInstances',
        field='managedInstances',
    )
    instances_to_return = []
    for instance_ref in managed_instances:
      if path_simplifier.Name(
          instance_ref.instance
      ) in instance_names or instance_ref.instance in instance_names:
        instances_to_return.append(instance_ref.instance)
    return instances_to_return
  else:
    raise ValueError('Unknown reference type {0}'.format(igm_ref.Collection()))


def GetDeviceNamesFromStatefulPolicy(stateful_policy):
  """Returns a list of device names from given StatefulPolicy message."""
  if (not stateful_policy or not stateful_policy.preservedState
      or not stateful_policy.preservedState.disks):
    return []
  return [disk.key
          for disk in stateful_policy.preservedState.disks.additionalProperties]


def GetInterfaceNamesFromStatefulPolicyForInternalIPs(stateful_policy):
  """Returns a list of stateful internal IPs interface names."""
  return _GetInterfaceNamesFromStatefulPolicyForIPs(
      stateful_policy,
      lambda stateful_policy: stateful_policy.preservedState.internalIPs)


def GetInterfaceNamesFromStatefulPolicyForExternalIPs(stateful_policy):
  """Returns a list of stateful external IPs interface names."""
  return _GetInterfaceNamesFromStatefulPolicyForIPs(
      stateful_policy,
      lambda stateful_policy: stateful_policy.preservedState.externalIPs)


def _GetInterfaceNamesFromStatefulPolicyForIPs(stateful_policy,
                                               ips_field_lambda):
  if not stateful_policy or not stateful_policy.preservedState:
    return []
  ips = ips_field_lambda(stateful_policy)
  if not ips:
    return []
  return [ip.key for ip in ips.additionalProperties]


def IsAutoscalerNew(autoscaler):
  return getattr(autoscaler, 'name', None) is None


def PatchUpdatePolicy(client, args, update_policy):
  """Returns an update_policy with attributes from args applied."""
  result = None
  if update_policy is None:
    update_policy = client.messages.InstanceGroupManagerUpdatePolicy()
  if args.IsSpecified('instance_redistribution_type'):
    update_policy.instanceRedistributionType = (
        client.messages.InstanceGroupManagerUpdatePolicy
        .InstanceRedistributionTypeValueValuesEnum)(
            args.instance_redistribution_type.upper())
    result = update_policy
  if args.IsSpecified('update_policy_type'):
    update_policy.type = update_instances_utils.ParseUpdatePolicyType(
        '--update-policy-update-type', args.update_policy_type, client.messages)
    result = update_policy
  if args.IsSpecified('update_policy_max_surge'):
    update_policy.maxSurge = update_instances_utils.ParseFixedOrPercent(
        '--update-policy-max-surge', 'max-surge', args.update_policy_max_surge,
        client.messages)
    result = update_policy
  if args.IsSpecified('update_policy_max_unavailable'):
    update_policy.maxUnavailable = update_instances_utils.ParseFixedOrPercent(
        '--update-policy-max-unavailable', 'max-unavailable',
        args.update_policy_max_unavailable, client.messages)
    result = update_policy
  if args.IsSpecified('update_policy_minimal_action'):
    update_policy.minimalAction = (
        update_instances_utils.ParseInstanceActionFlag)(
            '--update-policy-minimal-action', args.update_policy_minimal_action,
            client.messages.InstanceGroupManagerUpdatePolicy
            .MinimalActionValueValuesEnum)
    result = update_policy
  if args.IsSpecified('update_policy_most_disruptive_action'):
    update_policy.mostDisruptiveAllowedAction = (
        update_instances_utils.ParseInstanceActionFlag)(
            '--update-policy-most-disruptive-action',
            args.update_policy_most_disruptive_action,
            client.messages.InstanceGroupManagerUpdatePolicy
            .MostDisruptiveAllowedActionValueValuesEnum)
    result = update_policy
  if args.IsSpecified('update_policy_replacement_method'):
    replacement_method = update_instances_utils.ParseReplacementMethod(
        args.update_policy_replacement_method, client.messages)
    update_policy.replacementMethod = replacement_method
    result = update_policy
  # min_ready is available in alpha and beta APIs only
  if hasattr(args, 'update_policy_min_ready') and args.IsSpecified(
      'update_policy_min_ready'):
    update_policy.minReadySec = args.update_policy_min_ready
    result = update_policy
  return result


def ListPerInstanceConfigs(client, igm_ref):
  """Lists per-instance-configs for a given IGM."""
  if not hasattr(client.messages,
                 'ComputeInstanceGroupManagersListPerInstanceConfigsRequest'):
    return []
  if igm_ref.Collection() == 'compute.instanceGroupManagers':
    service = client.apitools_client.instanceGroupManagers
    request = (client.messages
               .ComputeInstanceGroupManagersListPerInstanceConfigsRequest)(
                   instanceGroupManager=igm_ref.Name(),
                   project=igm_ref.project,
                   zone=igm_ref.zone,
               )
  elif igm_ref.Collection() == 'compute.regionInstanceGroupManagers':
    service = client.apitools_client.regionInstanceGroupManagers
    request = (client.messages.
               ComputeRegionInstanceGroupManagersListPerInstanceConfigsRequest)(
                   instanceGroupManager=igm_ref.Name(),
                   project=igm_ref.project,
                   region=igm_ref.region,
               )
  else:
    raise ValueError('Unknown reference type {0}'.format(igm_ref.Collection()))

  errors = []
  results = list(
      request_helper.MakeRequests(
          requests=[(service, 'ListPerInstanceConfigs', request)],
          http=client.apitools_client.http,
          batch_url=client.batch_url,
          errors=errors))

  if not results:
    return []
  return results[0].items


def IsStateful(client, igm_info, igm_ref):
  """For a given IGM, returns if it is stateful."""
  # TODO(b/129752312): use igm.is_stateful field when launched
  pics = ListPerInstanceConfigs(client, igm_ref)
  has_policy = (
      hasattr(igm_info, 'statefulPolicy') and
      igm_info.statefulPolicy is not None)

  return has_policy or pics


def ValidateIgmReadyForStatefulness(igm_resource, client):
  """Throws exception if IGM is in state not ready for adding statefulness."""
  if not igm_resource.updatePolicy:
    return

  client_update_policy = client.messages.InstanceGroupManagerUpdatePolicy
  type_is_proactive = (
      igm_resource.updatePolicy.type == (
          client_update_policy.TypeValueValuesEnum.PROACTIVE))
  replacement_method_is_substitute = (
      igm_resource.updatePolicy.replacementMethod == (
          client_update_policy.ReplacementMethodValueValuesEnum.SUBSTITUTE))
  instance_redistribution_type_is_proactive = (
      igm_resource.updatePolicy.instanceRedistributionType == (
          client_update_policy.InstanceRedistributionTypeValueValuesEnum
          .PROACTIVE))

  if type_is_proactive and replacement_method_is_substitute:
    raise exceptions.Error(
        'Stateful IGMs cannot use SUBSTITUTE replacement method. '
        'Try `gcloud compute instance-groups managed '
        'rolling-update stop-proactive-update')
  if instance_redistribution_type_is_proactive:
    raise exceptions.Error(
        'Stateful regional IGMs cannot use proactive instance redistribution. '
        'Try `gcloud compute instance-groups managed '
        'update --instance-redistribution-type=NONE')


def ValueOrNone(message):
  """Return message if message is a proto with one or more fields set or None.

  If message is None or is the default proto, it returns None. In all other
  cases, it returns the message.

  Args:
    message: An generated proto message object.

  Returns:
    message if message is initialized or None
  """
  if message is None:
    return message
  default_object = message.__class__()
  return message if message != default_object else None


def RegisterCustomInstancePropertiesPatchEncoders(client):
  """Registers decoders and encoders that will handle null values for maps in InstancePropertiesPatch message."""

  def _NullValueEncoder(message):
    """Encoder for use when removing map entries.

    It works around issues with proto encoding of AdditionalProperties with null
    values by directly encoding a dict of keys with None values into json,
    skipping proto-based encoding.

    Args:
      message: an instance of InstancePropertiesPatch.MetadataValue or
        InstancePropertiesPatch.LabelsValue

    Returns:
      JSON string with null value.
    """
    return json.dumps({
        property.key: property.value
        for property in message.additionalProperties
    })

  def _NullMetadataDecoder(data):
    """Decoder for metadata map entries.

    Args:
      data: JSON representation of metadata map.

    Returns:
      Instance of InstancePropertiesPatch.MetadataValue.
    """
    py_object = json.loads(data)
    return client.messages.InstancePropertiesPatch.MetadataValue(
        additionalProperties=[
            client.messages.InstancePropertiesPatch.MetadataValue
            .AdditionalProperty(key=key, value=value)
            for key, value in py_object.items()
        ])

  def _NullLabelsDecoder(data):
    """Decoder for labels map entries.

    Args:
      data: JSON representation of labels map.

    Returns:
      Instance of InstancePropertiesPatch.LabelsValue.
    """
    py_object = json.loads(data)
    return client.messages.InstancePropertiesPatch.LabelsValue(
        additionalProperties=[
            client.messages.InstancePropertiesPatch.LabelsValue
            .AdditionalProperty(key=key, value=value)
            for key, value in py_object.items()
        ])

  encoding.RegisterCustomMessageCodec(
      encoder=_NullValueEncoder, decoder=_NullMetadataDecoder)(
          client.messages.InstancePropertiesPatch.MetadataValue)
  encoding.RegisterCustomMessageCodec(
      encoder=_NullValueEncoder, decoder=_NullLabelsDecoder)(
          client.messages.InstancePropertiesPatch.LabelsValue)


def RegisterCustomStatefulIpsPatchEncoders(client):
  """Registers decoders and encoders that will handle null values for Internal and External IPs maps in StatefulPolicy message."""

  auto_delete_map = {
      'NEVER':
          client.messages.StatefulPolicyPreservedStateNetworkIp(
              autoDelete=client.messages.StatefulPolicyPreservedStateNetworkIp
              .AutoDeleteValueValuesEnum.NEVER),
      'ON_PERMANENT_INSTANCE_DELETION':
          client.messages.StatefulPolicyPreservedStateNetworkIp(
              autoDelete=client.messages.StatefulPolicyPreservedStateNetworkIp
              .AutoDeleteValueValuesEnum.ON_PERMANENT_INSTANCE_DELETION)
  }

  def _StatefulIpsValueEncoder(message):
    """Encoder for Stateful Ips map entries.

    It works around issues with proto encoding of StatefulPolicyPreservedState
    with null values by directly encoding a dict of keys with None values into
    json, skipping proto-based encoding.

    Args:
      message: an instance of StatefulPolicyPreservedState.InternalIPsValue or
        StatefulPolicyPreservedState.ExternalIPsValue

    Returns:
      JSON string with null value.
    """
    return json.dumps({
        property.key: _GetAutodeleteOrNone(property)
        for property in message.additionalProperties
    })

  def _GetAutodeleteOrNone(autodelete):
    if autodelete.value is None:
      return None
    return {'autoDelete': autodelete.value.autoDelete.name}

  def _InternalStatefulIpsDecoder(data):
    """Decoder for Stateful Ips map entries.

    Args:
      data: JSON representation of Internal Stateful Ips.

    Returns:
      Instance of StatefulPolicyPreservedState.InternalIPsValue.
    """
    ips_value = client.messages.StatefulPolicyPreservedState.InternalIPsValue
    return _CommonStatefulIpsDecoder(data, ips_value)

  def _ExternalStatefulIpsDecoder(data):
    """Decoder for Stateful Ips map entries.

    Args:
      data: JSON representation of External Stateful Ips.

    Returns:
      Instance of
      StatefulPolicyPreservedState.ExternalIPsValue.AdditionalProperty.
    """
    ips_value = client.messages.StatefulPolicyPreservedState.ExternalIPsValue
    return _CommonStatefulIpsDecoder(data, ips_value)

  def _CommonStatefulIpsDecoder(data, ips_value):
    py_object = json.loads(data)
    return ips_value(additionalProperties=[
        ips_value.AdditionalProperty(
            key=key, value=auto_delete_map[value['autoDelete']])
        for key, value in py_object.items()
    ])

  encoding.RegisterCustomMessageCodec(
      encoder=_StatefulIpsValueEncoder, decoder=_InternalStatefulIpsDecoder)(
          client.messages.StatefulPolicyPreservedState.InternalIPsValue)
  encoding.RegisterCustomMessageCodec(
      encoder=_StatefulIpsValueEncoder, decoder=_ExternalStatefulIpsDecoder)(
          client.messages.StatefulPolicyPreservedState.ExternalIPsValue)


def RegisterCustomStatefulDisksPatchEncoders(client):
  """Registers decoders and encoders that will handle null values for Stateful Disks map in StatefulPolicy message."""

  auto_delete_map = {
      'NEVER':
          client.messages.StatefulPolicyPreservedStateDiskDevice(
              autoDelete=client.messages.StatefulPolicyPreservedStateDiskDevice
              .AutoDeleteValueValuesEnum.NEVER),
      'ON_PERMANENT_INSTANCE_DELETION':
          client.messages.StatefulPolicyPreservedStateDiskDevice(
              autoDelete=client.messages.StatefulPolicyPreservedStateDiskDevice
              .AutoDeleteValueValuesEnum.ON_PERMANENT_INSTANCE_DELETION)
  }

  def _StatefulDisksValueEncoder(message):
    """Encoder for Stateful Disks map entries.

    It works around issues with proto encoding of StatefulPolicyPreservedState
    with null values by directly encoding a dict of keys with None values into
    json, skipping proto-based encoding.

    Args:
      message: an instance of StatefulPolicyPreservedState.DisksValue

    Returns:
      JSON string with null value.
    """
    return json.dumps({
        property.key: _GetAutodeleteOrNone(property)
        for property in message.additionalProperties
    })

  def _GetAutodeleteOrNone(autodelete):
    if autodelete.value is None:
      return None
    return {'autoDelete': autodelete.value.autoDelete.name}

  def _StatefulDisksDecoder(data):
    """Decoder for Stateful Disks map entries.

    Args:
      data: JSON representation of Stateful Disks.

    Returns:
      Instance of StatefulPolicyPreservedState.DisksValue.
    """
    disk_device_value = client.messages.StatefulPolicyPreservedState.DisksValue
    py_object = json.loads(data)
    return disk_device_value(additionalProperties=[
        disk_device_value.AdditionalProperty(
            key=key, value=auto_delete_map[value['autoDelete']])
        for key, value in py_object.items()
    ])

  encoding.RegisterCustomMessageCodec(
      encoder=_StatefulDisksValueEncoder, decoder=_StatefulDisksDecoder)(
          client.messages.StatefulPolicyPreservedState.DisksValue)



"""
    pygments.formatters
    ~~~~~~~~~~~~~~~~~~~

    Pygments formatters.

    :copyright: Copyright 2006-2022 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import re
import sys
import types
import fnmatch
from os.path import basename

from pygments.formatters._mapping import FORMATTERS
from pygments.plugin import find_plugin_formatters
from pygments.util import ClassNotFound

__all__ = ['get_formatter_by_name', 'get_formatter_for_filename',
           'get_all_formatters', 'load_formatter_from_file'] + list(FORMATTERS)

_formatter_cache = {}  # classes by name
_pattern_cache = {}


def _fn_matches(fn, glob):
    """Return whether the supplied file name fn matches pattern filename."""
    if glob not in _pattern_cache:
        pattern = _pattern_cache[glob] = re.compile(fnmatch.translate(glob))
        return pattern.match(fn)
    return _pattern_cache[glob].match(fn)


def _load_formatters(module_name):
    """Load a formatter (and all others in the module too)."""
    mod = __import__(module_name, None, None, ['__all__'])
    for formatter_name in mod.__all__:
        cls = getattr(mod, formatter_name)
        _formatter_cache[cls.name] = cls


def get_all_formatters():
    """Return a generator for all formatter classes."""
    # NB: this returns formatter classes, not info like get_all_lexers().
    for info in FORMATTERS.values():
        if info[1] not in _formatter_cache:
            _load_formatters(info[0])
        yield _formatter_cache[info[1]]
    for _, formatter in find_plugin_formatters():
        yield formatter


def find_formatter_class(alias):
    """Lookup a formatter by alias.

    Returns None if not found.
    """
    for module_name, name, aliases, _, _ in FORMATTERS.values():
        if alias in aliases:
            if name not in _formatter_cache:
                _load_formatters(module_name)
            return _formatter_cache[name]
    for _, cls in find_plugin_formatters():
        if alias in cls.aliases:
            return cls


def get_formatter_by_name(_alias, **options):
    """Lookup and instantiate a formatter by alias.

    Raises ClassNotFound if not found.
    """
    cls = find_formatter_class(_alias)
    if cls is None:
        raise ClassNotFound("no formatter found for name %r" % _alias)
    return cls(**options)


def load_formatter_from_file(filename, formattername="CustomFormatter",
                             **options):
    """Load a formatter from a file.

    This method expects a file located relative to the current working
    directory, which contains a class named CustomFormatter. By default,
    it expects the Formatter to be named CustomFormatter; you can specify
    your own class name as the second argument to this function.

    Users should be very careful with the input, because this method
    is equivalent to running eval on the input file.

    Raises ClassNotFound if there are any problems importing the Formatter.

    .. versionadded:: 2.2
    """
    try:
        # This empty dict will contain the namespace for the exec'd file
        custom_namespace = {}
        with open(filename, 'rb') as f:
            exec(f.read(), custom_namespace)
        # Retrieve the class `formattername` from that namespace
        if formattername not in custom_namespace:
            raise ClassNotFound('no valid %s class found in %s' %
                                (formattername, filename))
        formatter_class = custom_namespace[formattername]
        # And finally instantiate it with the options
        return formatter_class(**options)
    except OSError as err:
        raise ClassNotFound('cannot read %s: %s' % (filename, err))
    except ClassNotFound:
        raise
    except Exception as err:
        raise ClassNotFound('error when loading custom formatter: %s' % err)


def get_formatter_for_filename(fn, **options):
    """Lookup and instantiate a formatter by filename pattern.

    Raises ClassNotFound if not found.
    """
    fn = basename(fn)
    for modname, name, _, filenames, _ in FORMATTERS.values():
        for filename in filenames:
            if _fn_matches(fn, filename):
                if name not in _formatter_cache:
                    _load_formatters(modname)
                return _formatter_cache[name](**options)
    for cls in find_plugin_formatters():
        for filename in cls.filenames:
            if _fn_matches(fn, filename):
                return cls(**options)
    raise ClassNotFound("no formatter found for file name %r" % fn)


class _automodule(types.ModuleType):
    """Automatically import formatters."""

    def __getattr__(self, name):
        info = FORMATTERS.get(name)
        if info:
            _load_formatters(info[0])
            cls = _formatter_cache[info[1]]
            setattr(self, name, cls)
            return cls
        raise AttributeError(name)


oldmod = sys.modules[__name__]
newmod = _automodule(__name__)
newmod.__dict__.update(oldmod.__dict__)
sys.modules[__name__] = newmod
del newmod.newmod, newmod.oldmod, newmod.sys, newmod.types
